# -*- coding: utf-8 -*-
"""
wsproto/events
~~~~~~~~~~~~~~

Events that result from processing data on a WebSocket connection.
"""


class Event(object):
    """
    Base class for wsproto events.
    """
    pass


class ConnectionRequested(Event):
    """
    The ConnectionRequested event is fired when a SERVER connection receives a
    WebSocket handshake request (HTTP with upgrade header).
    """
    def __init__(self, proposed_subprotocols, h11request):
        self.proposed_subprotocols = proposed_subprotocols
        self.h11request = h11request

    def __repr__(self):
        path = self.h11request.target

        headers = dict(self.h11request.headers)
        host = headers[b'host']
        version = headers[b'sec-websocket-version']
        subprotocol = headers.get(b'sec-websocket-protocol', None)
        extensions = []

        fmt = '<%s host=%s path=%s version=%s subprotocol=%r extensions=%r>'
        return fmt % (self.__class__.__name__, host, path, version,
                      subprotocol, extensions)


class ConnectionEstablished(Event):
    """
    The ConnectionEstablished event is fired when a CLIENT connection completes
    the WebSocket handshake and is ready to send & receive messages.
    """
    def __init__(self, subprotocol=None, extensions=None):
        self.subprotocol = subprotocol
        self.extensions = extensions
        if self.extensions is None:
            self.extensions = []

    def __repr__(self):
        return '<ConnectionEstablished subprotocol=%r extensions=%r>' % \
               (self.subprotocol, self.extensions)


class ConnectionClosed(Event):
    """
    The ConnectionClosed event is fired after the connection is considered closed.

    wsproto automatically emits a CLOSE frame when it receives one, to complete the close-handshake.
    """
    def __init__(self, code, reason=None):
        #: The close status code, see :class:`CloseReason
        #: <wsproto.frame_protocol.CloseReason>`.
        self.code = code
        self.reason = reason

    def __repr__(self):
        return '<%s code=%r reason="%s">' % (self.__class__.__name__,
                                             self.code, self.reason)


class ConnectionFailed(ConnectionClosed):
    pass


class DataReceived(Event):
    def __init__(self, data, frame_finished, message_finished):
        #: The message data as byte string, can be decoded as UTF-8 for TEXT messages.
        #: This only represents a single chunk of data and not a full WebSocket message.
        #: You need to buffer and reassemble these chunks to get the full message.
        self.data = data

        #: This has no semantic content, but is provided just in case some
        #: weird edge case user wants to be able to reconstruct the
        #: fragmentation pattern of the original stream. You don't want it:
        self.frame_finished = frame_finished

        #: True if this frame is the last one of this message, False if more frames are expected.
        self.message_finished = message_finished


class TextReceived(DataReceived):
    """
    The TextReceived event is fired when a data frame with TEXT payload is received.
    """
    pass


class BytesReceived(DataReceived):
    """
    The BytesReceived event is fired when a data frame with BINARY payload is received.
    """
    pass


class PingReceived(Event):
    """
    The PingReceived event is fired when a Ping is received.

    wsproto automatically emits a PONG frame with the same payload.
    """
    def __init__(self, payload):
        #: Optional "Application data", i.e., binary payload.
        self.payload = payload


class PongReceived(Event):
    """
    The PongReceived event is fired when a Pong is received.
    """
    def __init__(self, payload):
        #: Optional "Application data", i.e., binary payload.
        #: Make sure to verify against the orignal PING payload.
        self.payload = payload



# flake8: noqa

import sys


PY2 = sys.version_info.major == 2
PY3 = sys.version_info.major == 3


if PY3:
    unicode = str

    def Utf8Validator():
        return None
else:
    unicode = unicode
    try:
        from wsaccel.utf8validator import Utf8Validator
    except ImportError:
        from .utf8validator import Utf8Validator



import select
import socket

from wsproto.connection import WSConnection, SERVER, ConnectionRequested, \
                               ConnectionClosed
from wsproto.events import DataReceived
from wsproto.extensions import PerMessageDeflate

count = 0

def new_conn(sock):
    global count
    print("test_server.py received connection {}".format(count))
    count += 1
    ws = WSConnection(SERVER, extensions=[PerMessageDeflate()])
    closed = False
    while not closed:
        try:
            data = sock.recv(65535)
        except socket.error:
            data = None

        ws.receive_bytes(data or None)

        for event in ws.events():
            if isinstance(event, ConnectionRequested):
                ws.accept(event)
            elif isinstance(event, DataReceived):
                ws.send_data(event.data, event.message_finished)
            elif isinstance(event, ConnectionClosed):
                closed = True

        if not data:
            closed = True

        try:
            data = ws.bytes_to_send()
            sock.sendall(data)
        except socket.error:
            closed = True

    sock.close()

def start_listener(host='127.0.0.1', port=8642, shutdown_port=8643):
    server = socket.socket()
    server.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
    server.bind((host, port))
    server.listen(1)
    shutdown_server = socket.socket()
    shutdown_server.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
    shutdown_server.bind((host, shutdown_port))
    shutdown_server.listen(1)

    done = False
    filenos = {s.fileno(): s for s in (server, shutdown_server)}

    while not done:
        r, _, _ = select.select(filenos.keys(), [], [], 0)

        for sock in [filenos[fd] for fd in r]:
            if sock is server:
                new_conn(server.accept()[0])
            else:
                done = True

if __name__ == '__main__':
    try:
        start_listener()
    except KeyboardInterrupt:
        pass



# -*- coding: utf-8 -*-

import itertools
from binascii import unhexlify
from codecs import getincrementaldecoder
import struct

import pytest

import wsproto.frame_protocol as fp
import wsproto.extensions as wpext


class FakeValidator(object):
    def __init__(self):
        self.validated = b''

        self.valid = True
        self.ends_on_complete = True
        self.octet = 0
        self.code_point = 0

    def validate(self, data):
        self.validated += data
        return (self.valid, self.ends_on_complete, self.octet,
                self.code_point)


class TestBuffer(object):
    def test_consume_at_most_zero_bytes(self):
        buf = fp.Buffer(b'xxyyy')
        assert buf.consume_at_most(0) == bytearray()

    def test_consume_at_most_with_no_data(self):
        buf = fp.Buffer()
        assert buf.consume_at_most(1) == bytearray()

    def test_consume_at_most_with_sufficient_data(self):
        buf = fp.Buffer(b'xx')
        assert buf.consume_at_most(2) == b'xx'

    def test_consume_at_most_with_more_than_sufficient_data(self):
        buf = fp.Buffer(b'xxyyy')
        assert buf.consume_at_most(2) == b'xx'

    def test_consume_at_most_with_insufficient_data(self):
        buf = fp.Buffer(b'xx')
        assert buf.consume_at_most(3) == b'xx'

    def test_consume_exactly_with_sufficient_data(self):
        buf = fp.Buffer(b'xx')
        assert buf.consume_exactly(2) == b'xx'

    def test_consume_exactly_with_more_than_sufficient_data(self):
        buf = fp.Buffer(b'xxyyy')
        assert buf.consume_exactly(2) == b'xx'

    def test_consume_exactly_with_insufficient_data(self):
        buf = fp.Buffer(b'xx')
        assert buf.consume_exactly(3) is None

    def test_feed(self):
        buf = fp.Buffer()
        assert buf.consume_at_most(1) == b''
        assert buf.consume_exactly(1) is None
        buf.feed(b'xy')
        assert buf.consume_at_most(1) == b'x'
        assert buf.consume_exactly(1) == b'y'

    def test_rollback(self):
        buf = fp.Buffer()
        buf.feed(b'xyz')
        assert buf.consume_exactly(2) == b'xy'
        assert buf.consume_exactly(1) == b'z'
        assert buf.consume_at_most(1) == b''
        buf.rollback()
        assert buf.consume_at_most(3) == b'xyz'

    def test_commit(self):
        buf = fp.Buffer()
        buf.feed(b'xyz')
        assert buf.consume_exactly(2) == b'xy'
        assert buf.consume_exactly(1) == b'z'
        assert buf.consume_at_most(1) == b''
        buf.commit()
        assert buf.consume_at_most(3) == b''

    def test_length(self):
        buf = fp.Buffer()
        data = b'xyzabc'
        buf.feed(data)
        assert len(buf) == len(data)


class TestMessageDecoder(object):
    def test_single_binary_frame(self):
        payload = b'x' * 23
        decoder = fp.MessageDecoder()
        frame = fp.Frame(
            opcode=fp.Opcode.BINARY,
            payload=payload,
            frame_finished=True,
            message_finished=True,
        )

        frame = decoder.process_frame(frame)
        assert frame.opcode is fp.Opcode.BINARY
        assert frame.message_finished is True
        assert frame.payload == payload

    def test_follow_on_binary_frame(self):
        payload = b'x' * 23
        decoder = fp.MessageDecoder()
        decoder.opcode = fp.Opcode.BINARY
        decoder.seen_first_frame = True
        frame = fp.Frame(
            opcode=fp.Opcode.CONTINUATION,
            payload=payload,
            frame_finished=True,
            message_finished=False,
        )

        frame = decoder.process_frame(frame)
        assert frame.opcode is fp.Opcode.BINARY
        assert frame.message_finished is False
        assert frame.payload == payload

    def test_single_text_frame(self):
        text_payload = u'fñör∂'
        binary_payload = text_payload.encode('utf8')
        decoder = fp.MessageDecoder()
        frame = fp.Frame(
            opcode=fp.Opcode.TEXT,
            payload=binary_payload,
            frame_finished=True,
            message_finished=True,
        )

        frame = decoder.process_frame(frame)
        assert frame.opcode is fp.Opcode.TEXT
        assert frame.message_finished is True
        assert frame.payload == text_payload

    def test_follow_on_text_frame(self):
        text_payload = u'fñör∂'
        binary_payload = text_payload.encode('utf8')
        decoder = fp.MessageDecoder()
        decoder.opcode = fp.Opcode.TEXT
        decoder.seen_first_frame = True
        decoder.decoder = getincrementaldecoder("utf-8")()

        assert decoder.decoder.decode(binary_payload[:4]) == text_payload[:2]
        binary_payload = binary_payload[4:-2]
        text_payload = text_payload[2:-1]

        frame = fp.Frame(
            opcode=fp.Opcode.CONTINUATION,
            payload=binary_payload,
            frame_finished=True,
            message_finished=False,
        )

        frame = decoder.process_frame(frame)
        assert frame.opcode is fp.Opcode.TEXT
        assert frame.message_finished is False
        assert frame.payload == text_payload

    def test_final_text_frame(self):
        text_payload = u'fñör∂'
        binary_payload = text_payload.encode('utf8')
        decoder = fp.MessageDecoder()
        decoder.opcode = fp.Opcode.TEXT
        decoder.seen_first_frame = True
        decoder.decoder = getincrementaldecoder("utf-8")()

        assert decoder.decoder.decode(binary_payload[:-2]) == text_payload[:-1]
        binary_payload = binary_payload[-2:]
        text_payload = text_payload[-1:]

        frame = fp.Frame(
            opcode=fp.Opcode.CONTINUATION,
            payload=binary_payload,
            frame_finished=True,
            message_finished=True,
        )

        frame = decoder.process_frame(frame)
        assert frame.opcode is fp.Opcode.TEXT
        assert frame.message_finished is True
        assert frame.payload == text_payload

    def test_start_with_continuation(self):
        payload = b'x' * 23
        decoder = fp.MessageDecoder()
        frame = fp.Frame(
            opcode=fp.Opcode.CONTINUATION,
            payload=payload,
            frame_finished=True,
            message_finished=True,
        )

        with pytest.raises(fp.ParseFailed):
            decoder.process_frame(frame)

    def test_missing_continuation_1(self):
        payload = b'x' * 23
        decoder = fp.MessageDecoder()
        decoder.opcode = fp.Opcode.BINARY
        decoder.seen_first_frame = True
        frame = fp.Frame(
            opcode=fp.Opcode.BINARY,
            payload=payload,
            frame_finished=True,
            message_finished=True,
        )

        with pytest.raises(fp.ParseFailed):
            decoder.process_frame(frame)

    def test_missing_continuation_2(self):
        payload = b'x' * 23
        decoder = fp.MessageDecoder()
        decoder.opcode = fp.Opcode.TEXT
        frame = fp.Frame(
            opcode=fp.Opcode.BINARY,
            payload=payload,
            frame_finished=True,
            message_finished=True,
        )

        with pytest.raises(fp.ParseFailed):
            decoder.process_frame(frame)

    def test_incomplete_unicode(self):
        payload = u'fñör∂'
        payload = payload.encode('utf8')
        payload = payload[:4]

        decoder = fp.MessageDecoder()
        frame = fp.Frame(
            opcode=fp.Opcode.TEXT,
            payload=payload,
            frame_finished=True,
            message_finished=True,
        )

        with pytest.raises(fp.ParseFailed) as excinfo:
            decoder.process_frame(frame)
        assert excinfo.value.code is fp.CloseReason.INVALID_FRAME_PAYLOAD_DATA

    def test_not_even_unicode(self):
        payload = u'fñörd'
        payload = payload.encode('iso-8859-1')

        decoder = fp.MessageDecoder()
        frame = fp.Frame(
            opcode=fp.Opcode.TEXT,
            payload=payload,
            frame_finished=True,
            message_finished=False,
        )

        with pytest.raises(fp.ParseFailed) as excinfo:
            decoder.process_frame(frame)
        assert excinfo.value.code is fp.CloseReason.INVALID_FRAME_PAYLOAD_DATA

    def test_bad_unicode(self):
        payload = unhexlify('cebae1bdb9cf83cebcceb5eda080656469746564')

        decoder = fp.MessageDecoder()
        frame = fp.Frame(
            opcode=fp.Opcode.TEXT,
            payload=payload,
            frame_finished=True,
            message_finished=True,
        )

        with pytest.raises(fp.ParseFailed) as excinfo:
            decoder.process_frame(frame)
        assert excinfo.value.code is fp.CloseReason.INVALID_FRAME_PAYLOAD_DATA

    def test_split_message(self):
        text_payload = u'x' * 65535
        payload = text_payload.encode('utf-8')
        split = 32777

        decoder = fp.MessageDecoder()

        frame = fp.Frame(
            opcode=fp.Opcode.TEXT,
            payload=payload[:split],
            frame_finished=False,
            message_finished=True
        )
        frame = decoder.process_frame(frame)
        assert frame.opcode is fp.Opcode.TEXT
        assert frame.message_finished is False
        assert frame.payload == text_payload[:split]

        frame = fp.Frame(
            opcode=fp.Opcode.CONTINUATION,
            payload=payload[split:],
            frame_finished=True,
            message_finished=True
        )
        frame = decoder.process_frame(frame)
        assert frame.opcode is fp.Opcode.TEXT
        assert frame.message_finished is True
        assert frame.payload == text_payload[split:]

    def test_split_unicode_message(self):
        text_payload = u'∂' * 64
        payload = text_payload.encode('utf-8')
        split = 64

        decoder = fp.MessageDecoder()

        frame = fp.Frame(
            opcode=fp.Opcode.TEXT,
            payload=payload[:split],
            frame_finished=False,
            message_finished=True
        )
        frame = decoder.process_frame(frame)
        assert frame.opcode is fp.Opcode.TEXT
        assert frame.message_finished is False
        assert frame.payload == text_payload[:(split // 3)]

        frame = fp.Frame(
            opcode=fp.Opcode.CONTINUATION,
            payload=payload[split:],
            frame_finished=True,
            message_finished=True
        )
        frame = decoder.process_frame(frame)
        assert frame.opcode is fp.Opcode.TEXT
        assert frame.message_finished is True
        assert frame.payload == text_payload[(split // 3):]

    def send_frame_to_validator(self, payload, finished):
        decoder = fp.MessageDecoder()
        frame = fp.Frame(
            opcode=fp.Opcode.TEXT,
            payload=payload,
            frame_finished=finished,
            message_finished=True,
        )
        frame = decoder.process_frame(frame)

    def test_text_message_hits_validator(self, monkeypatch):
        validator = FakeValidator()
        monkeypatch.setattr(fp, 'Utf8Validator', lambda: validator)

        text_payload = u'fñör∂'
        binary_payload = text_payload.encode('utf8')
        self.send_frame_to_validator(binary_payload, True)

        assert validator.validated == binary_payload

    def test_message_validation_failure_fails_properly(self, monkeypatch):
        validator = FakeValidator()
        validator.valid = False

        monkeypatch.setattr(fp, 'Utf8Validator', lambda: validator)

        with pytest.raises(fp.ParseFailed):
            self.send_frame_to_validator(b'', True)

    def test_message_validation_finish_on_incomplete(self, monkeypatch):
        validator = FakeValidator()
        validator.valid = True
        validator.ends_on_complete = False

        monkeypatch.setattr(fp, 'Utf8Validator', lambda: validator)

        with pytest.raises(fp.ParseFailed):
            self.send_frame_to_validator(b'', True)

    def test_message_validation_unfinished_on_incomplete(self, monkeypatch):
        validator = FakeValidator()
        validator.valid = True
        validator.ends_on_complete = False

        monkeypatch.setattr(fp, 'Utf8Validator', lambda: validator)

        self.send_frame_to_validator(b'', False)

    def test_message_no_validation_can_still_fail(self, monkeypatch):
        validator = FakeValidator()
        monkeypatch.setattr(fp, 'Utf8Validator', lambda: validator)

        payload = u'fñörd'
        payload = payload.encode('iso-8859-1')

        with pytest.raises(fp.ParseFailed) as exc:
            self.send_frame_to_validator(payload, True)
        assert exc.value.code == fp.CloseReason.INVALID_FRAME_PAYLOAD_DATA


class TestFrameDecoder(object):
    def _single_frame_test(self, client, frame_bytes, opcode, payload,
                           frame_finished, message_finished):
        decoder = fp.FrameDecoder(client=client)
        decoder.receive_bytes(frame_bytes)
        frame = decoder.process_buffer()
        assert frame is not None
        assert frame.opcode is opcode
        assert frame.payload == payload
        assert frame.frame_finished is frame_finished
        assert frame.message_finished is message_finished

    def _split_frame_test(self, client, frame_bytes, opcode, payload,
                          frame_finished, message_finished, split):
        decoder = fp.FrameDecoder(client=client)
        decoder.receive_bytes(frame_bytes[:split])
        assert decoder.process_buffer() is None
        decoder.receive_bytes(frame_bytes[split:])
        frame = decoder.process_buffer()
        assert frame is not None
        assert frame.opcode is opcode
        assert frame.payload == payload
        assert frame.frame_finished is frame_finished
        assert frame.message_finished is message_finished

    def _split_message_test(self, client, frame_bytes, opcode, payload, split):
        decoder = fp.FrameDecoder(client=client)

        decoder.receive_bytes(frame_bytes[:split])
        frame = decoder.process_buffer()
        assert frame is not None
        assert frame.opcode is opcode
        assert frame.payload == payload[:len(frame.payload)]
        assert frame.frame_finished is False
        assert frame.message_finished is True

        decoder.receive_bytes(frame_bytes[split:])
        frame = decoder.process_buffer()
        assert frame is not None
        assert frame.opcode is fp.Opcode.CONTINUATION
        assert frame.payload == payload[-len(frame.payload):]
        assert frame.frame_finished is True
        assert frame.message_finished is True

    def _parse_failure_test(self, client, frame_bytes, close_reason):
        decoder = fp.FrameDecoder(client=client)
        with pytest.raises(fp.ParseFailed) as excinfo:
            decoder.receive_bytes(frame_bytes)
            decoder.process_buffer()
        assert excinfo.value.code is close_reason

    def test_zero_length_message(self):
        self._single_frame_test(
            client=True,
            frame_bytes=b'\x81\x00',
            opcode=fp.Opcode.TEXT,
            payload=b'',
            frame_finished=True,
            message_finished=True,
        )

    def test_short_server_message_frame(self):
        self._single_frame_test(
            client=True,
            frame_bytes=b'\x81\x02xy',
            opcode=fp.Opcode.TEXT,
            payload=b'xy',
            frame_finished=True,
            message_finished=True,
        )

    def test_short_client_message_frame(self):
        self._single_frame_test(
            client=False,
            frame_bytes=b'\x81\x82abcd\x19\x1b',
            opcode=fp.Opcode.TEXT,
            payload=b'xy',
            frame_finished=True,
            message_finished=True,
        )

    def test_reject_masked_server_frame(self):
        self._parse_failure_test(
            client=True,
            frame_bytes=b'\x81\x82abcd\x19\x1b',
            close_reason=fp.CloseReason.PROTOCOL_ERROR,
        )

    def test_reject_unmasked_client_frame(self):
        self._parse_failure_test(
            client=False,
            frame_bytes=b'\x81\x02xy',
            close_reason=fp.CloseReason.PROTOCOL_ERROR,
        )

    def test_reject_bad_opcode(self):
        self._parse_failure_test(
            client=True,
            frame_bytes=b'\x8e\x02xy',
            close_reason=fp.CloseReason.PROTOCOL_ERROR,
        )

    def test_reject_unfinished_control_frame(self):
        self._parse_failure_test(
            client=True,
            frame_bytes=b'\x09\x02xy',
            close_reason=fp.CloseReason.PROTOCOL_ERROR,
        )

    def test_reject_reserved_bits(self):
        self._parse_failure_test(
            client=True,
            frame_bytes=b'\x91\x02xy',
            close_reason=fp.CloseReason.PROTOCOL_ERROR,
        )
        self._parse_failure_test(
            client=True,
            frame_bytes=b'\xa1\x02xy',
            close_reason=fp.CloseReason.PROTOCOL_ERROR,
        )
        self._parse_failure_test(
            client=True,
            frame_bytes=b'\xc1\x02xy',
            close_reason=fp.CloseReason.PROTOCOL_ERROR,
        )

    def test_long_message_frame(self):
        payload = b'x' * 512
        payload_len = struct.pack('!H', len(payload))
        frame_bytes = b'\x81\x7e' + payload_len + payload

        self._single_frame_test(
            client=True,
            frame_bytes=frame_bytes,
            opcode=fp.Opcode.TEXT,
            payload=payload,
            frame_finished=True,
            message_finished=True,
        )

    def test_very_long_message_frame(self):
        payload = b'x' * (128 * 1024)
        payload_len = struct.pack('!Q', len(payload))
        frame_bytes = b'\x81\x7f' + payload_len + payload

        self._single_frame_test(
            client=True,
            frame_bytes=frame_bytes,
            opcode=fp.Opcode.TEXT,
            payload=payload,
            frame_finished=True,
            message_finished=True,
        )

    def test_insufficiently_long_message_frame(self):
        payload = b'x' * 64
        payload_len = struct.pack('!H', len(payload))
        frame_bytes = b'\x81\x7e' + payload_len + payload

        self._parse_failure_test(
            client=True,
            frame_bytes=frame_bytes,
            close_reason=fp.CloseReason.PROTOCOL_ERROR,
        )

    def test_insufficiently_very_long_message_frame(self):
        payload = b'x' * 512
        payload_len = struct.pack('!Q', len(payload))
        frame_bytes = b'\x81\x7f' + payload_len + payload

        self._parse_failure_test(
            client=True,
            frame_bytes=frame_bytes,
            close_reason=fp.CloseReason.PROTOCOL_ERROR,
        )

    def test_very_insufficiently_very_long_message_frame(self):
        payload = b'x' * 64
        payload_len = struct.pack('!Q', len(payload))
        frame_bytes = b'\x81\x7f' + payload_len + payload

        self._parse_failure_test(
            client=True,
            frame_bytes=frame_bytes,
            close_reason=fp.CloseReason.PROTOCOL_ERROR,
        )

    def test_not_enough_for_header(self):
        payload = b'xy'
        frame_bytes = b'\x81\x02' + payload

        self._split_frame_test(
            client=True,
            frame_bytes=frame_bytes,
            opcode=fp.Opcode.TEXT,
            payload=payload,
            frame_finished=True,
            message_finished=True,
            split=1,
        )

    def test_not_enough_for_long_length(self):
        payload = b'x' * 512
        payload_len = struct.pack('!H', len(payload))
        frame_bytes = b'\x81\x7e' + payload_len + payload

        self._split_frame_test(
            client=True,
            frame_bytes=frame_bytes,
            opcode=fp.Opcode.TEXT,
            payload=payload,
            frame_finished=True,
            message_finished=True,
            split=3,
        )

    def test_not_enough_for_very_long_length(self):
        payload = b'x' * (128 * 1024)
        payload_len = struct.pack('!Q', len(payload))
        frame_bytes = b'\x81\x7f' + payload_len + payload

        self._split_frame_test(
            client=True,
            frame_bytes=frame_bytes,
            opcode=fp.Opcode.TEXT,
            payload=payload,
            frame_finished=True,
            message_finished=True,
            split=7,
        )

    def test_eight_byte_length_with_msb_set(self):
        frame_bytes = b'\x81\x7f\x80\x80\x80\x80\x80\x80\x80\x80'

        self._parse_failure_test(
            client=True,
            frame_bytes=frame_bytes,
            close_reason=fp.CloseReason.PROTOCOL_ERROR,
        )

    def test_not_enough_for_mask(self):
        payload = bytearray(b'xy')
        mask = bytearray(b'abcd')
        masked_payload = bytearray([
            payload[0] ^ mask[0],
            payload[1] ^ mask[1]
        ])
        frame_bytes = b'\x81\x82' + mask + masked_payload

        self._split_frame_test(
            client=False,
            frame_bytes=frame_bytes,
            opcode=fp.Opcode.TEXT,
            payload=payload,
            frame_finished=True,
            message_finished=True,
            split=4,
        )

    def test_partial_message_frames(self):
        chunk_size = 1024
        payload = b'x' * (128 * chunk_size)
        payload_len = struct.pack('!Q', len(payload))
        frame_bytes = b'\x81\x7f' + payload_len + payload
        header_len = len(frame_bytes) - len(payload)

        decoder = fp.FrameDecoder(client=True)
        decoder.receive_bytes(frame_bytes[:header_len])
        assert decoder.process_buffer() is None
        frame_bytes = frame_bytes[header_len:]
        payload_sent = 0
        expected_opcode = fp.Opcode.TEXT
        for offset in range(0, len(frame_bytes), chunk_size):
            chunk = frame_bytes[offset:offset + chunk_size]
            decoder.receive_bytes(chunk)
            frame = decoder.process_buffer()
            payload_sent += chunk_size
            all_payload_sent = payload_sent == len(payload)
            assert frame is not None
            assert frame.opcode is expected_opcode
            assert frame.frame_finished is all_payload_sent
            assert frame.message_finished is True
            assert frame.payload == payload[offset:offset + chunk_size]

            expected_opcode = fp.Opcode.CONTINUATION

    def test_partial_control_frame(self):
        chunk_size = 11
        payload = b'x' * 64
        frame_bytes = b'\x89' + bytearray([len(payload)]) + payload

        decoder = fp.FrameDecoder(client=True)

        for offset in range(0, len(frame_bytes) - chunk_size, chunk_size):
            chunk = frame_bytes[offset:offset + chunk_size]
            decoder.receive_bytes(chunk)
            assert decoder.process_buffer() is None

        decoder.receive_bytes(frame_bytes[-chunk_size:])
        frame = decoder.process_buffer()
        assert frame is not None
        assert frame.opcode is fp.Opcode.PING
        assert frame.frame_finished is True
        assert frame.message_finished is True
        assert frame.payload == payload

    def test_long_message_sliced(self):
        payload = b'x' * 65535
        payload_len = struct.pack('!H', len(payload))
        frame_bytes = b'\x81\x7e' + payload_len + payload

        self._split_message_test(
            client=True,
            frame_bytes=frame_bytes,
            opcode=fp.Opcode.TEXT,
            payload=payload,
            split=65535,
        )

    def test_overly_long_control_frame(self):
        payload = b'x' * 128
        payload_len = struct.pack('!H', len(payload))
        frame_bytes = b'\x89\x7e' + payload_len + payload

        self._parse_failure_test(
            client=True,
            frame_bytes=frame_bytes,
            close_reason=fp.CloseReason.PROTOCOL_ERROR,
        )


class TestFrameDecoderExtensions(object):
    class FakeExtension(wpext.Extension):
        name = 'fake'

        def __init__(self):
            self._inbound_header_called = False
            self._inbound_rsv_bit_set = False
            self._inbound_payload_data_called = False
            self._inbound_complete_called = False
            self._fail_inbound_complete = False
            self._outbound_rsv_bit_set = False

        def enabled(self):
            return True

        def frame_inbound_header(self, proto, opcode, rsv, payload_length):
            self._inbound_header_called = True
            if opcode is fp.Opcode.PONG:
                return fp.CloseReason.MANDATORY_EXT
            self._inbound_rsv_bit_set = rsv.rsv3
            return fp.RsvBits(False, False, True)

        def frame_inbound_payload_data(self, proto, data):
            self._inbound_payload_data_called = True
            if data == b'party time':
                return fp.CloseReason.POLICY_VIOLATION
            elif data == b'ragequit':
                self._fail_inbound_complete = True
            if self._inbound_rsv_bit_set:
                data = data.decode('utf-8').upper().encode('utf-8')
            return data

        def frame_inbound_complete(self, proto, fin):
            self._inbound_complete_called = True
            if self._fail_inbound_complete:
                return fp.CloseReason.ABNORMAL_CLOSURE
            if fin and self._inbound_rsv_bit_set:
                return u'™'.encode('utf-8')

        def frame_outbound(self, proto, opcode, rsv, data, fin):
            if opcode is fp.Opcode.TEXT:
                rsv = fp.RsvBits(rsv.rsv1, rsv.rsv2, True)
                self._outbound_rsv_bit_set = True
            if fin and self._outbound_rsv_bit_set:
                data += u'®'.encode('utf-8')
                self._outbound_rsv_bit_set = False
            return rsv, data

    def test_rsv_bit(self):
        ext = self.FakeExtension()
        decoder = fp.FrameDecoder(client=True, extensions=[ext])

        frame_bytes = b'\x91\x00'

        decoder.receive_bytes(frame_bytes)
        frame = decoder.process_buffer()
        assert frame is not None
        assert ext._inbound_header_called
        assert ext._inbound_rsv_bit_set

    def test_wrong_rsv_bit(self):
        ext = self.FakeExtension()
        decoder = fp.FrameDecoder(client=True, extensions=[ext])

        frame_bytes = b'\xa1\x00'

        decoder.receive_bytes(frame_bytes)
        with pytest.raises(fp.ParseFailed) as excinfo:
            decoder.receive_bytes(frame_bytes)
            decoder.process_buffer()
        assert excinfo.value.code is fp.CloseReason.PROTOCOL_ERROR

    def test_header_error_handling(self):
        ext = self.FakeExtension()
        decoder = fp.FrameDecoder(client=True, extensions=[ext])

        frame_bytes = b'\x9a\x00'

        decoder.receive_bytes(frame_bytes)
        with pytest.raises(fp.ParseFailed) as excinfo:
            decoder.receive_bytes(frame_bytes)
            decoder.process_buffer()
        assert excinfo.value.code is fp.CloseReason.MANDATORY_EXT

    def test_payload_processing(self):
        ext = self.FakeExtension()
        decoder = fp.FrameDecoder(client=True, extensions=[ext])

        payload = u'fñör∂'
        expected_payload = payload.upper().encode('utf-8')
        bytes_payload = payload.encode('utf-8')
        frame_bytes = b'\x11' + bytearray([len(bytes_payload)]) + bytes_payload

        decoder.receive_bytes(frame_bytes)
        frame = decoder.process_buffer()
        assert frame is not None
        assert ext._inbound_header_called
        assert ext._inbound_rsv_bit_set
        assert ext._inbound_payload_data_called
        assert frame.payload == expected_payload

    def test_no_payload_processing_when_not_wanted(self):
        ext = self.FakeExtension()
        decoder = fp.FrameDecoder(client=True, extensions=[ext])

        payload = u'fñör∂'
        expected_payload = payload.encode('utf-8')
        bytes_payload = payload.encode('utf-8')
        frame_bytes = b'\x01' + bytearray([len(bytes_payload)]) + bytes_payload

        decoder.receive_bytes(frame_bytes)
        frame = decoder.process_buffer()
        assert frame is not None
        assert ext._inbound_header_called
        assert not ext._inbound_rsv_bit_set
        assert ext._inbound_payload_data_called
        assert frame.payload == expected_payload

    def test_payload_error_handling(self):
        ext = self.FakeExtension()
        decoder = fp.FrameDecoder(client=True, extensions=[ext])

        payload = b'party time'
        frame_bytes = b'\x91' + bytearray([len(payload)]) + payload

        decoder.receive_bytes(frame_bytes)
        with pytest.raises(fp.ParseFailed) as excinfo:
            decoder.receive_bytes(frame_bytes)
            decoder.process_buffer()
        assert excinfo.value.code is fp.CloseReason.POLICY_VIOLATION

    def test_frame_completion(self):
        ext = self.FakeExtension()
        decoder = fp.FrameDecoder(client=True, extensions=[ext])

        payload = u'fñör∂'
        expected_payload = (payload + u'™').upper().encode('utf-8')
        bytes_payload = payload.encode('utf-8')
        frame_bytes = b'\x91' + bytearray([len(bytes_payload)]) + bytes_payload

        decoder.receive_bytes(frame_bytes)
        frame = decoder.process_buffer()
        assert frame is not None
        assert ext._inbound_header_called
        assert ext._inbound_rsv_bit_set
        assert ext._inbound_payload_data_called
        assert ext._inbound_complete_called
        assert frame.payload == expected_payload

    def test_no_frame_completion_when_not_wanted(self):
        ext = self.FakeExtension()
        decoder = fp.FrameDecoder(client=True, extensions=[ext])

        payload = u'fñör∂'
        expected_payload = payload.encode('utf-8')
        bytes_payload = payload.encode('utf-8')
        frame_bytes = b'\x81' + bytearray([len(bytes_payload)]) + bytes_payload

        decoder.receive_bytes(frame_bytes)
        frame = decoder.process_buffer()
        assert frame is not None
        assert ext._inbound_header_called
        assert not ext._inbound_rsv_bit_set
        assert ext._inbound_payload_data_called
        assert ext._inbound_complete_called
        assert frame.payload == expected_payload

    def test_completion_error_handling(self):
        ext = self.FakeExtension()
        decoder = fp.FrameDecoder(client=True, extensions=[ext])

        payload = b'ragequit'
        frame_bytes = b'\x91' + bytearray([len(payload)]) + payload

        decoder.receive_bytes(frame_bytes)
        with pytest.raises(fp.ParseFailed) as excinfo:
            decoder.receive_bytes(frame_bytes)
            decoder.process_buffer()
        assert excinfo.value.code is fp.CloseReason.ABNORMAL_CLOSURE

    def test_outbound_handling_single_frame(self):
        ext = self.FakeExtension()
        proto = fp.FrameProtocol(client=False, extensions=[ext])
        payload = u'😃😄🙃😉'
        data = proto.send_data(payload, fin=True)
        payload = (payload + u'®').encode('utf8')
        assert data == b'\x91' + bytearray([len(payload)]) + payload

    def test_outbound_handling_multiple_frames(self):
        ext = self.FakeExtension()
        proto = fp.FrameProtocol(client=False, extensions=[ext])
        payload = u'😃😄🙃😉'
        data = proto.send_data(payload, fin=False)
        payload = payload.encode('utf8')
        assert data == b'\x11' + bytearray([len(payload)]) + payload

        payload = u'¯\_(ツ)_/¯'
        data = proto.send_data(payload, fin=True)
        payload = (payload + u'®').encode('utf8')
        assert data == b'\x80' + bytearray([len(payload)]) + payload


class TestFrameProtocolReceive(object):
    def test_long_text_message(self):
        payload = 'x' * 65535
        encoded_payload = payload.encode('utf-8')
        payload_len = struct.pack('!H', len(encoded_payload))
        frame_bytes = b'\x81\x7e' + payload_len + encoded_payload

        protocol = fp.FrameProtocol(client=True, extensions=[])
        protocol.receive_bytes(frame_bytes)
        frames = list(protocol.received_frames())
        assert len(frames) == 1
        frame = frames[0]
        assert frame.opcode == fp.Opcode.TEXT
        assert len(frame.payload) == len(payload)
        assert frame.payload == payload

    def _close_test(self, code, reason=None, reason_bytes=None):
        payload = b''
        if code:
            payload += struct.pack('!H', code)
        if reason:
            payload += reason.encode('utf8')
        elif reason_bytes:
            payload += reason_bytes

        frame_bytes = b'\x88' + bytearray([len(payload)]) + payload

        protocol = fp.FrameProtocol(client=True, extensions=[])
        protocol.receive_bytes(frame_bytes)
        frames = list(protocol.received_frames())
        assert len(frames) == 1
        frame = frames[0]
        assert frame.opcode == fp.Opcode.CLOSE
        assert frame.payload[0] == code or fp.CloseReason.NO_STATUS_RCVD
        if reason:
            assert frame.payload[1] == reason
        else:
            assert not frame.payload[1]

    def test_close_no_code(self):
        self._close_test(None)

    def test_close_one_byte_code(self):
        frame_bytes = b'\x88\x01\x0e'
        protocol = fp.FrameProtocol(client=True, extensions=[])

        with pytest.raises(fp.ParseFailed) as exc:
            protocol.receive_bytes(frame_bytes)
            list(protocol.received_frames())
        assert exc.value.code == fp.CloseReason.PROTOCOL_ERROR

    def test_close_bad_code(self):
        with pytest.raises(fp.ParseFailed) as exc:
            self._close_test(123)
        assert exc.value.code == fp.CloseReason.PROTOCOL_ERROR

    def test_close_unknown_code(self):
        with pytest.raises(fp.ParseFailed) as exc:
            self._close_test(2998)
        assert exc.value.code == fp.CloseReason.PROTOCOL_ERROR

    def test_close_local_only_code(self):
        with pytest.raises(fp.ParseFailed) as exc:
            self._close_test(fp.CloseReason.NO_STATUS_RCVD)
        assert exc.value.code == fp.CloseReason.PROTOCOL_ERROR

    def test_close_no_payload(self):
        self._close_test(fp.CloseReason.NORMAL_CLOSURE)

    def test_close_easy_payload(self):
        self._close_test(fp.CloseReason.NORMAL_CLOSURE, u'tarah old chap')

    def test_close_utf8_payload(self):
        self._close_test(fp.CloseReason.NORMAL_CLOSURE, u'fñør∂')

    def test_close_bad_utf8_payload(self):
        payload = unhexlify('cebae1bdb9cf83cebcceb5eda080656469746564')
        with pytest.raises(fp.ParseFailed) as exc:
            self._close_test(fp.CloseReason.NORMAL_CLOSURE,
                             reason_bytes=payload)
        assert exc.value.code == fp.CloseReason.INVALID_FRAME_PAYLOAD_DATA

    def test_close_incomplete_utf8_payload(self):
        payload = u'fñør∂'.encode('utf8')[:-1]
        with pytest.raises(fp.ParseFailed) as exc:
            self._close_test(fp.CloseReason.NORMAL_CLOSURE,
                             reason_bytes=payload)
        assert exc.value.code == fp.CloseReason.INVALID_FRAME_PAYLOAD_DATA

    def test_random_control_frame(self):
        payload = b'give me one ping vasily'
        frame_bytes = b'\x89' + bytearray([len(payload)]) + payload

        protocol = fp.FrameProtocol(client=True, extensions=[])
        protocol.receive_bytes(frame_bytes)
        frames = list(protocol.received_frames())
        assert len(frames) == 1
        frame = frames[0]
        assert frame.opcode == fp.Opcode.PING
        assert len(frame.payload) == len(payload)
        assert frame.payload == payload

    def test_close_reasons_get_utf8_validated(self, monkeypatch):
        validator = FakeValidator()
        reason = u'ƒñø®∂'
        monkeypatch.setattr(fp, 'Utf8Validator', lambda: validator)

        self._close_test(fp.CloseReason.NORMAL_CLOSURE, reason=reason)

        assert validator.validated == reason.encode('utf8')

    def test_close_reason_failing_validation_fails(self, monkeypatch):
        validator = FakeValidator()
        validator.valid = False
        reason = u'ƒñø®∂'
        monkeypatch.setattr(fp, 'Utf8Validator', lambda: validator)

        with pytest.raises(fp.ParseFailed) as exc:
            self._close_test(fp.CloseReason.NORMAL_CLOSURE, reason=reason)
        assert exc.value.code == fp.CloseReason.INVALID_FRAME_PAYLOAD_DATA

    def test_close_reason_with_incomplete_utf8_fails(self, monkeypatch):
        validator = FakeValidator()
        validator.ends_on_complete = False
        reason = u'ƒñø®∂'
        monkeypatch.setattr(fp, 'Utf8Validator', lambda: validator)

        with pytest.raises(fp.ParseFailed) as exc:
            self._close_test(fp.CloseReason.NORMAL_CLOSURE, reason=reason)
        assert exc.value.code == fp.CloseReason.INVALID_FRAME_PAYLOAD_DATA

    def test_close_no_validation(self, monkeypatch):
        monkeypatch.setattr(fp, 'Utf8Validator', lambda: None)
        reason = u'ƒñø®∂'
        self._close_test(fp.CloseReason.NORMAL_CLOSURE, reason=reason)

    def test_close_no_validation_can_still_fail(self, monkeypatch):
        validator = FakeValidator()
        monkeypatch.setattr(fp, 'Utf8Validator', lambda: validator)

        reason = u'fñörd'
        reason = reason.encode('iso-8859-1')

        with pytest.raises(fp.ParseFailed) as exc:
            self._close_test(fp.CloseReason.NORMAL_CLOSURE,
                             reason_bytes=reason)
        assert exc.value.code == fp.CloseReason.INVALID_FRAME_PAYLOAD_DATA


class TestFrameProtocolSend(object):
    def test_simplest_possible_close(self):
        proto = fp.FrameProtocol(client=False, extensions=[])
        data = proto.close()
        assert data == b'\x88\x00'

    def test_unreasoning_close(self):
        proto = fp.FrameProtocol(client=False, extensions=[])
        data = proto.close(code=fp.CloseReason.NORMAL_CLOSURE)
        assert data == b'\x88\x02\x03\xe8'

    def test_reasoned_close(self):
        proto = fp.FrameProtocol(client=False, extensions=[])
        reason = u'¯\_(ツ)_/¯'
        expected_payload = struct.pack('!H', fp.CloseReason.NORMAL_CLOSURE) + \
            reason.encode('utf8')
        data = proto.close(code=fp.CloseReason.NORMAL_CLOSURE, reason=reason)
        assert data == b'\x88' + bytearray([len(expected_payload)]) + \
            expected_payload

    def test_overly_reasoned_close(self):
        proto = fp.FrameProtocol(client=False, extensions=[])
        reason = u'¯\_(ツ)_/¯' * 10
        data = proto.close(code=fp.CloseReason.NORMAL_CLOSURE, reason=reason)
        assert bytes(data[0:1]) == b'\x88'
        assert len(data) <= 127
        assert data[4:].decode('utf8')

    def test_reasoned_but_uncoded_close(self):
        proto = fp.FrameProtocol(client=False, extensions=[])
        with pytest.raises(TypeError):
            proto.close(reason='termites')

    def test_local_only_close_reason(self):
        proto = fp.FrameProtocol(client=False, extensions=[])
        data = proto.close(code=fp.CloseReason.NO_STATUS_RCVD)
        assert data == b'\x88\x02\x03\xe8'

    def test_ping_without_payload(self):
        proto = fp.FrameProtocol(client=False, extensions=[])
        data = proto.ping()
        assert data == b'\x89\x00'

    def test_ping_with_payload(self):
        proto = fp.FrameProtocol(client=False, extensions=[])
        payload = u'¯\_(ツ)_/¯'.encode('utf8')
        data = proto.ping(payload)
        assert data == b'\x89' + bytearray([len(payload)]) + payload

    def test_pong_without_payload(self):
        proto = fp.FrameProtocol(client=False, extensions=[])
        data = proto.pong()
        assert data == b'\x8a\x00'

    def test_pong_with_payload(self):
        proto = fp.FrameProtocol(client=False, extensions=[])
        payload = u'¯\_(ツ)_/¯'.encode('utf8')
        data = proto.pong(payload)
        assert data == b'\x8a' + bytearray([len(payload)]) + payload

    def test_single_short_binary_data(self):
        proto = fp.FrameProtocol(client=False, extensions=[])
        payload = b"it's all just ascii, right?"
        data = proto.send_data(payload, fin=True)
        assert data == b'\x82' + bytearray([len(payload)]) + payload

    def test_single_short_text_data(self):
        proto = fp.FrameProtocol(client=False, extensions=[])
        payload = u'😃😄🙃😉'
        data = proto.send_data(payload, fin=True)
        payload = payload.encode('utf8')
        assert data == b'\x81' + bytearray([len(payload)]) + payload

    def test_multiple_short_binary_data(self):
        proto = fp.FrameProtocol(client=False, extensions=[])
        payload = b"it's all just ascii, right?"
        data = proto.send_data(payload, fin=False)
        assert data == b'\x02' + bytearray([len(payload)]) + payload

        payload = b'sure no worries'
        data = proto.send_data(payload, fin=True)
        assert data == b'\x80' + bytearray([len(payload)]) + payload

    def test_multiple_short_text_data(self):
        proto = fp.FrameProtocol(client=False, extensions=[])
        payload = u'😃😄🙃😉'
        data = proto.send_data(payload, fin=False)
        payload = payload.encode('utf8')
        assert data == b'\x01' + bytearray([len(payload)]) + payload

        payload = u'🙈🙉🙊'
        data = proto.send_data(payload, fin=True)
        payload = payload.encode('utf8')
        assert data == b'\x80' + bytearray([len(payload)]) + payload

    def test_mismatched_data_messages1(self):
        proto = fp.FrameProtocol(client=False, extensions=[])
        payload = u'😃😄🙃😉'
        data = proto.send_data(payload, fin=False)
        payload = payload.encode('utf8')
        assert data == b'\x01' + bytearray([len(payload)]) + payload

        payload = b'seriously, all ascii'
        with pytest.raises(TypeError):
            proto.send_data(payload)

    def test_mismatched_data_messages2(self):
        proto = fp.FrameProtocol(client=False, extensions=[])
        payload = b"it's all just ascii, right?"
        data = proto.send_data(payload, fin=False)
        assert data == b'\x02' + bytearray([len(payload)]) + payload

        payload = u'✔️☑️✅✔︎☑'
        with pytest.raises(TypeError):
            proto.send_data(payload)

    def test_message_length_max_short(self):
        proto = fp.FrameProtocol(client=False, extensions=[])
        payload = b'x' * 125
        data = proto.send_data(payload, fin=True)
        assert data == b'\x82' + bytearray([len(payload)]) + payload

    def test_message_length_min_two_byte(self):
        proto = fp.FrameProtocol(client=False, extensions=[])
        payload = b'x' * 126
        data = proto.send_data(payload, fin=True)
        assert data == b'\x82\x7e' + struct.pack('!H', len(payload)) + payload

    def test_message_length_max_two_byte(self):
        proto = fp.FrameProtocol(client=False, extensions=[])
        payload = b'x' * (2 ** 16 - 1)
        data = proto.send_data(payload, fin=True)
        assert data == b'\x82\x7e' + struct.pack('!H', len(payload)) + payload

    def test_message_length_min_eight_byte(self):
        proto = fp.FrameProtocol(client=False, extensions=[])
        payload = b'x' * (2 ** 16)
        data = proto.send_data(payload, fin=True)
        assert data == b'\x82\x7f' + struct.pack('!Q', len(payload)) + payload

    def test_client_side_masking_short_frame(self):
        proto = fp.FrameProtocol(client=True, extensions=[])
        payload = b'x' * 125
        data = proto.send_data(payload, fin=True)
        assert data[0] == 0x82
        assert struct.unpack('!B', data[1:2])[0] == len(payload) | 0x80
        masking_key = data[2:6]
        maskbytes = itertools.cycle(masking_key)
        assert data[6:] == \
            bytearray(b ^ next(maskbytes) for b in bytearray(payload))

    def test_client_side_masking_two_byte_frame(self):
        proto = fp.FrameProtocol(client=True, extensions=[])
        payload = b'x' * 126
        data = proto.send_data(payload, fin=True)
        assert data[0] == 0x82
        assert data[1] == 0xfe
        assert struct.unpack('!H', data[2:4])[0] == len(payload)
        masking_key = data[4:8]
        maskbytes = itertools.cycle(masking_key)
        assert data[8:] == \
            bytearray(b ^ next(maskbytes) for b in bytearray(payload))

    def test_client_side_masking_eight_byte_frame(self):
        proto = fp.FrameProtocol(client=True, extensions=[])
        payload = b'x' * 65536
        data = proto.send_data(payload, fin=True)
        assert data[0] == 0x82
        assert data[1] == 0xff
        assert struct.unpack('!Q', data[2:10])[0] == len(payload)
        masking_key = data[10:14]
        maskbytes = itertools.cycle(masking_key)
        assert data[14:] == \
            bytearray(b ^ next(maskbytes) for b in bytearray(payload))

    def test_control_frame_with_overly_long_payload(self):
        proto = fp.FrameProtocol(client=False, extensions=[])
        payload = b'x' * 126

        with pytest.raises(ValueError):
            proto.pong(payload)

    def test_data_we_have_no_idea_what_to_do_with(self):
        proto = fp.FrameProtocol(client=False, extensions=[])
        payload = dict()

        with pytest.raises(ValueError):
            proto.send_data(payload)



# -*- coding: utf-8 -*-

import zlib

import pytest

import wsproto.extensions as wpext
import wsproto.frame_protocol as fp


class TestPerMessageDeflate(object):
    parameter_sets = [
        {
            'client_no_context_takeover': False,
            'client_max_window_bits': 15,
            'server_no_context_takeover': False,
            'server_max_window_bits': 15,
        },
        {
            'client_no_context_takeover': True,
            'client_max_window_bits': 9,
            'server_no_context_takeover': False,
            'server_max_window_bits': 15,
        },
        {
            'client_no_context_takeover': False,
            'client_max_window_bits': 15,
            'server_no_context_takeover': True,
            'server_max_window_bits': 9,
        },
        {
            'client_no_context_takeover': True,
            'client_max_window_bits': 8,
            'server_no_context_takeover': True,
            'server_max_window_bits': 9,
        },
        {
            'client_no_context_takeover': True,
            'server_max_window_bits': 9,
        },
        {
            'server_no_context_takeover': True,
            'client_max_window_bits': 8,
        },
        {
            'client_max_window_bits': None,
            'server_max_window_bits': None,
        },
        {},
    ]

    def make_offer_string(self, params):
        offer = ['permessage-deflate']

        if 'client_max_window_bits' in params:
            if params['client_max_window_bits'] is None:
                offer.append('client_max_window_bits')
            else:
                offer.append('client_max_window_bits=%d' %
                             params['client_max_window_bits'])
        if 'server_max_window_bits' in params:
            if params['server_max_window_bits'] is None:
                offer.append('server_max_window_bits')
            else:
                offer.append('server_max_window_bits=%d' %
                             params['server_max_window_bits'])
        if params.get('client_no_context_takeover', False):
            offer.append('client_no_context_takeover')
        if params.get('server_no_context_takeover', False):
            offer.append('server_no_context_takeover')

        return '; '.join(offer)

    def compare_params_to_string(self, params, ext, param_string):
        if 'client_max_window_bits' in params:
            if params['client_max_window_bits'] is None:
                bits = ext.client_max_window_bits
            else:
                bits = params['client_max_window_bits']
            assert 'client_max_window_bits=%d' % bits in param_string
        if 'server_max_window_bits' in params:
            if params['server_max_window_bits'] is None:
                bits = ext.server_max_window_bits
            else:
                bits = params['server_max_window_bits']
            assert 'server_max_window_bits=%d' % bits in param_string
        if params.get('client_no_context_takeover', False):
            assert 'client_no_context_takeover' in param_string
        if params.get('server_no_context_takeover', False):
            assert 'server_no_context_takeover' in param_string

    @pytest.mark.parametrize('params', parameter_sets)
    def test_offer(self, params):
        ext = wpext.PerMessageDeflate(**params)
        offer = ext.offer(None)

        self.compare_params_to_string(params, ext, offer)

    @pytest.mark.parametrize('params', parameter_sets)
    def test_finalize(self, params):
        ext = wpext.PerMessageDeflate()
        assert not ext.enabled()

        params = dict(params)
        if 'client_max_window_bits' in params:
            if params['client_max_window_bits'] is None:
                del params['client_max_window_bits']
        if 'server_max_window_bits' in params:
            if params['server_max_window_bits'] is None:
                del params['server_max_window_bits']
        offer = self.make_offer_string(params)
        ext.finalize(None, offer)

        if params.get('client_max_window_bits', None):
            assert ext.client_max_window_bits == \
                params['client_max_window_bits']
        if params.get('server_max_window_bits', None):
            assert ext.server_max_window_bits == \
                params['server_max_window_bits']
        assert ext.client_no_context_takeover is \
            params.get('client_no_context_takeover', False)
        assert ext.server_no_context_takeover is \
            params.get('server_no_context_takeover', False)

        assert ext.enabled()

    def test_finalize_ignores_rubbish(self):
        ext = wpext.PerMessageDeflate()
        assert not ext.enabled()

        ext.finalize(None, 'i am the lizard queen; worship me')

        assert ext.enabled()

    @pytest.mark.parametrize('params', parameter_sets)
    def test_accept(self, params):
        ext = wpext.PerMessageDeflate()
        assert not ext.enabled()

        offer = self.make_offer_string(params)
        response = ext.accept(None, offer)

        if ext.client_no_context_takeover:
            assert 'client_no_context_takeover' in response
        if ext.server_no_context_takeover:
            assert 'server_no_context_takeover' in response
        if 'client_max_window_bits' in params:
            if params['client_max_window_bits'] is None:
                bits = ext.client_max_window_bits
            else:
                bits = params['client_max_window_bits']
            assert ext.client_max_window_bits == bits
            assert 'client_max_window_bits=%d' % bits in response
        if 'server_max_window_bits' in params:
            if params['server_max_window_bits'] is None:
                bits = ext.server_max_window_bits
            else:
                bits = params['server_max_window_bits']
            assert ext.server_max_window_bits == bits
            assert 'server_max_window_bits=%d' % bits in response

    def test_accept_ignores_rubbish(self):
        ext = wpext.PerMessageDeflate()
        assert not ext.enabled()

        ext.accept(None, 'i am the lizard queen; worship me')

        assert ext.enabled()

    def test_inbound_uncompressed_control_frame(self):
        payload = b'x' * 23

        ext = wpext.PerMessageDeflate()
        ext._enabled = True
        proto = fp.FrameProtocol(client=True, extensions=[ext])

        result = ext.frame_inbound_header(proto, fp.Opcode.PING,
                                          fp.RsvBits(False, False, False),
                                          len(payload))
        assert result.rsv1

        data = ext.frame_inbound_payload_data(proto, payload)
        assert data == payload

        assert ext.frame_inbound_complete(proto, True) is None

    def test_inbound_compressed_control_frame(self):
        payload = b'x' * 23

        ext = wpext.PerMessageDeflate()
        ext._enabled = True
        proto = fp.FrameProtocol(client=True, extensions=[ext])

        result = ext.frame_inbound_header(proto, fp.Opcode.PING,
                                          fp.RsvBits(True, False, False),
                                          len(payload))
        assert result == fp.CloseReason.PROTOCOL_ERROR

    def test_inbound_compressed_continuation_frame(self):
        payload = b'x' * 23

        ext = wpext.PerMessageDeflate()
        ext._enabled = True
        proto = fp.FrameProtocol(client=True, extensions=[ext])

        result = ext.frame_inbound_header(proto, fp.Opcode.CONTINUATION,
                                          fp.RsvBits(True, False, False),
                                          len(payload))
        assert result == fp.CloseReason.PROTOCOL_ERROR

    def test_inbound_uncompressed_data_frame(self):
        payload = b'x' * 23

        ext = wpext.PerMessageDeflate()
        ext._enabled = True
        proto = fp.FrameProtocol(client=True, extensions=[ext])

        result = ext.frame_inbound_header(proto, fp.Opcode.BINARY,
                                          fp.RsvBits(False, False, False),
                                          len(payload))
        assert result.rsv1

        data = ext.frame_inbound_payload_data(proto, payload)
        assert data == payload

        assert ext.frame_inbound_complete(proto, True) is None

    @pytest.mark.parametrize('client', [True, False])
    def test_client_inbound_compressed_single_data_frame(self, client):
        payload = b'x' * 23
        compressed_payload = b'\xaa\xa8\xc0\n\x00\x00'

        ext = wpext.PerMessageDeflate()
        ext._enabled = True
        proto = fp.FrameProtocol(client=client, extensions=[ext])

        result = ext.frame_inbound_header(proto, fp.Opcode.BINARY,
                                          fp.RsvBits(True, False, False),
                                          len(compressed_payload))
        assert result.rsv1

        data = ext.frame_inbound_payload_data(proto, compressed_payload)
        data += ext.frame_inbound_complete(proto, True)
        assert data == payload

    @pytest.mark.parametrize('client', [True, False])
    def test_client_inbound_compressed_multiple_data_frames(self, client):
        payload = b'x' * 23
        compressed_payload = b'\xaa\xa8\xc0\n\x00\x00'
        split = 3
        data = b''

        ext = wpext.PerMessageDeflate()
        ext._enabled = True
        proto = fp.FrameProtocol(client=client, extensions=[ext])

        result = ext.frame_inbound_header(proto, fp.Opcode.BINARY,
                                          fp.RsvBits(True, False, False),
                                          split)
        assert result.rsv1
        result = ext.frame_inbound_payload_data(proto,
                                                compressed_payload[:split])
        assert not isinstance(result, fp.CloseReason)
        data += result
        assert ext.frame_inbound_complete(proto, False) is None

        result = ext.frame_inbound_header(proto, fp.Opcode.CONTINUATION,
                                          fp.RsvBits(False, False, False),
                                          len(compressed_payload) - split)
        assert result.rsv1
        result = ext.frame_inbound_payload_data(proto,
                                                compressed_payload[split:])
        assert not isinstance(result, fp.CloseReason)
        data += result

        result = ext.frame_inbound_complete(proto, True)
        assert not isinstance(result, fp.CloseReason)
        data += result

        assert data == payload

    @pytest.mark.parametrize('client', [True, False])
    def test_client_decompress_after_uncompressible_frame(self, client):
        ext = wpext.PerMessageDeflate()
        ext._enabled = True
        proto = fp.FrameProtocol(client=client, extensions=[ext])

        # A PING frame
        result = ext.frame_inbound_header(proto, fp.Opcode.PING,
                                          fp.RsvBits(False, False, False),
                                          0)
        result = ext.frame_inbound_payload_data(proto, b'')
        assert not isinstance(result, fp.CloseReason)
        assert ext.frame_inbound_complete(proto, True) is None

        # A compressed TEXT frame
        payload = b'x' * 23
        compressed_payload = b'\xaa\xa8\xc0\n\x00\x00'

        result = ext.frame_inbound_header(proto, fp.Opcode.TEXT,
                                          fp.RsvBits(True, False, False),
                                          len(compressed_payload))
        assert result.rsv1
        result = ext.frame_inbound_payload_data(proto, compressed_payload)
        assert result == payload

        result = ext.frame_inbound_complete(proto, True)
        assert not isinstance(result, fp.CloseReason)

    def test_inbound_bad_zlib_payload(self):
        compressed_payload = b'x' * 23

        ext = wpext.PerMessageDeflate()
        ext._enabled = True
        proto = fp.FrameProtocol(client=True, extensions=[ext])

        result = ext.frame_inbound_header(proto, fp.Opcode.BINARY,
                                          fp.RsvBits(True, False, False),
                                          len(compressed_payload))
        assert result.rsv1
        result = ext.frame_inbound_payload_data(proto, compressed_payload)
        assert result is fp.CloseReason.INVALID_FRAME_PAYLOAD_DATA

    def test_inbound_bad_zlib_decoder_end_state(self, monkeypatch):
        compressed_payload = b'x' * 23

        ext = wpext.PerMessageDeflate()
        ext._enabled = True
        proto = fp.FrameProtocol(client=True, extensions=[ext])

        result = ext.frame_inbound_header(proto, fp.Opcode.BINARY,
                                          fp.RsvBits(True, False, False),
                                          len(compressed_payload))
        assert result.rsv1

        class FailDecompressor(object):
            def decompress(self, data):
                return b''

            def flush(self):
                raise zlib.error()

        monkeypatch.setattr(ext, '_decompressor', FailDecompressor())

        result = ext.frame_inbound_complete(proto, True)
        assert result is fp.CloseReason.INVALID_FRAME_PAYLOAD_DATA

    @pytest.mark.parametrize('client,no_context_takeover', [
        (True, True),
        (True, False),
        (False, True),
        (False, False),
    ])
    def test_decompressor_reset(self, client, no_context_takeover):
        if client:
            args = {'server_no_context_takeover': no_context_takeover}
        else:
            args = {'client_no_context_takeover': no_context_takeover}
        ext = wpext.PerMessageDeflate(**args)
        ext._enabled = True
        proto = fp.FrameProtocol(client=client, extensions=[ext])

        result = ext.frame_inbound_header(proto, fp.Opcode.BINARY,
                                          fp.RsvBits(True, False, False), 0)
        assert result.rsv1

        assert ext._decompressor is not None

        result = ext.frame_inbound_complete(proto, True)
        assert not isinstance(result, fp.CloseReason)

        if no_context_takeover:
            assert ext._decompressor is None
        else:
            assert ext._decompressor is not None

        result = ext.frame_inbound_header(proto, fp.Opcode.BINARY,
                                          fp.RsvBits(True, False, False), 0)
        assert result.rsv1

        assert ext._decompressor is not None

    def test_outbound_uncompressible_opcode(self):
        ext = wpext.PerMessageDeflate()
        ext._enabled = True
        proto = fp.FrameProtocol(client=True, extensions=[ext])

        rsv = fp.RsvBits(False, False, False)
        payload = b'x' * 23

        rsv, data = ext.frame_outbound(proto, fp.Opcode.PING, rsv, payload,
                                       True)

        assert rsv.rsv1 is False
        assert data == payload

    @pytest.mark.parametrize('client', [True, False])
    def test_outbound_compress_single_frame(self, client):
        ext = wpext.PerMessageDeflate()
        ext._enabled = True
        proto = fp.FrameProtocol(client=client, extensions=[ext])

        rsv = fp.RsvBits(False, False, False)
        payload = b'x' * 23
        compressed_payload = b'\xaa\xa8\xc0\n\x00\x00'

        rsv, data = ext.frame_outbound(proto, fp.Opcode.BINARY, rsv, payload,
                                       True)

        assert rsv.rsv1 is True
        assert data == compressed_payload

    @pytest.mark.parametrize('client', [True, False])
    def test_outbound_compress_multiple_frames(self, client):
        ext = wpext.PerMessageDeflate()
        ext._enabled = True
        proto = fp.FrameProtocol(client=client, extensions=[ext])

        rsv = fp.RsvBits(False, False, False)
        payload = b'x' * 23
        split = 12
        compressed_payload = b'\xaa\xa8\xc0\n\x00\x00'

        rsv, data = ext.frame_outbound(proto, fp.Opcode.BINARY, rsv,
                                       payload[:split], False)
        assert rsv.rsv1 is True

        rsv = fp.RsvBits(False, False, False)
        rsv, more_data = ext.frame_outbound(proto, fp.Opcode.CONTINUATION, rsv,
                                            payload[split:], True)
        assert rsv.rsv1 is False
        assert data + more_data == compressed_payload

    @pytest.mark.parametrize('client,no_context_takeover', [
        (True, True),
        (True, False),
        (False, True),
        (False, False),
    ])
    def test_compressor_reset(self, client, no_context_takeover):
        if client:
            args = {'client_no_context_takeover': no_context_takeover}
        else:
            args = {'server_no_context_takeover': no_context_takeover}
        ext = wpext.PerMessageDeflate(**args)
        ext._enabled = True
        proto = fp.FrameProtocol(client=client, extensions=[ext])
        rsv = fp.RsvBits(False, False, False)

        rsv, data = ext.frame_outbound(proto, fp.Opcode.BINARY, rsv, b'',
                                       False)
        assert rsv.rsv1 is True
        assert ext._compressor is not None

        rsv = fp.RsvBits(False, False, False)
        rsv, data = ext.frame_outbound(proto, fp.Opcode.CONTINUATION, rsv, b'',
                                       True)
        assert rsv.rsv1 is False
        if no_context_takeover:
            assert ext._compressor is None
        else:
            assert ext._compressor is not None

        rsv = fp.RsvBits(False, False, False)
        rsv, data = ext.frame_outbound(proto, fp.Opcode.BINARY, rsv, b'',
                                       False)
        assert rsv.rsv1 is True
        assert ext._compressor is not None

    @pytest.mark.parametrize('params', parameter_sets)
    def test_repr(self, params):
        ext = wpext.PerMessageDeflate(**params)
        self.compare_params_to_string(params, ext, repr(ext))



# -*- coding: utf-8 -*-

import itertools

import pytest

from wsproto.connection import WSConnection, CLIENT, SERVER, ConnectionState
from wsproto.events import (
    ConnectionClosed,
    ConnectionEstablished,
    ConnectionRequested,
    TextReceived,
    BytesReceived,
    PingReceived,
    PongReceived)
from wsproto.frame_protocol import CloseReason, FrameProtocol


class TestConnection(object):
    def create_connection(self):
        server = WSConnection(SERVER)
        client = WSConnection(CLIENT, host='localhost', resource='foo')

        server.receive_bytes(client.bytes_to_send())
        event = next(server.events())
        assert isinstance(event, ConnectionRequested)

        server.accept(event)
        client.receive_bytes(server.bytes_to_send())
        assert isinstance(next(client.events()), ConnectionEstablished)

        return client, server

    def test_negotiation(self):
        self.create_connection()

    def test_default_args(self):
        with pytest.raises(ValueError, match="Host must not be None"):
            WSConnection(CLIENT, resource='/ws')
        with pytest.raises(ValueError, match="Resource must not be None"):
            WSConnection(CLIENT, host='localhost')

    @pytest.mark.parametrize('as_client,final', [
        (True, True),
        (True, False),
        (False, True),
        (False, False)
    ])
    def test_send_and_receive(self, as_client, final):
        client, server = self.create_connection()
        if as_client:
            me = client
            them = server
        else:
            me = server
            them = client

        data = b'x' * 23

        me.send_data(data, final)
        them.receive_bytes(me.bytes_to_send())

        event = next(them.events())
        assert isinstance(event, BytesReceived)
        assert event.data == data
        assert event.message_finished is final

    @pytest.mark.parametrize('as_client,code,reason', [
        (True, CloseReason.NORMAL_CLOSURE, u'bye'),
        (True, CloseReason.GOING_AWAY, u'👋👋'),
        (False, CloseReason.NORMAL_CLOSURE, u'bye'),
        (False, CloseReason.GOING_AWAY, u'👋👋'),
    ])
    def test_close(self, as_client, code, reason):
        client, server = self.create_connection()
        if as_client:
            me = client
            them = server
        else:
            me = server
            them = client

        me.close(code, reason)
        them.receive_bytes(me.bytes_to_send())

        event = next(them.events())
        assert isinstance(event, ConnectionClosed)
        assert event.code is code
        assert event.reason == reason

    @pytest.mark.parametrize('swap', [
        True,
        False
    ])
    def test_normal_closure(self, swap):
        if swap:
            completor, initiator = self.create_connection()
        else:
            initiator, completor = self.create_connection()

        # initiator sends CLOSE to completor
        initiator.close()
        completor.receive_bytes(initiator.bytes_to_send())

        # completor emits ConnectionClosed
        assert isinstance(next(completor.events()), ConnectionClosed)

        # completor enters CLOSED state
        assert completor.closed
        with pytest.raises(StopIteration):
            next(completor.events())

        # completor sends CLOSE back to initiator
        initiator.receive_bytes(completor.bytes_to_send())

        # initiator emits ConnectionClosed
        assert isinstance(next(initiator.events()), ConnectionClosed)

        # initiator enters CLOSED state
        assert initiator.closed
        with pytest.raises(StopIteration):
            next(initiator.events())

        completor.ping()
        with pytest.raises(ValueError):
            initiator.receive_bytes(completor.bytes_to_send())

    def test_abnormal_closure(self):
        client, server = self.create_connection()

        for conn in (client, server):
            conn.receive_bytes(None)
            event = next(conn.events())
            assert isinstance(event, ConnectionClosed)
            assert event.code is CloseReason.ABNORMAL_CLOSURE
            assert conn.closed

    def test_bytes_send_all(self):
        connection = WSConnection(SERVER)
        connection._outgoing = b'fnord fnord'
        assert connection.bytes_to_send() == b'fnord fnord'
        assert connection.bytes_to_send() == b''

    def test_bytes_send_some(self):
        connection = WSConnection(SERVER)
        connection._outgoing = b'fnord fnord'
        assert connection.bytes_to_send(5) == b'fnord'
        assert connection.bytes_to_send() == b' fnord'

    @pytest.mark.parametrize('as_client', [True, False])
    def test_ping_pong(self, as_client):
        client, server = self.create_connection()
        if as_client:
            me = client
            them = server
        else:
            me = server
            them = client

        payload = b'x' * 23

        # Send a PING message
        me.ping(payload)
        wire_data = me.bytes_to_send()

        # Verify that the peer emits the PingReceive event with the correct
        # payload.
        them.receive_bytes(wire_data)
        event = next(them.events())
        assert isinstance(event, PingReceived)
        assert event.payload == payload
        with pytest.raises(StopIteration):
            repr(next(them.events()))

        # Let the peer send the automatic PONG message
        wire_data = them.bytes_to_send()
        assert wire_data[0] == 0x8a
        masked = bool(wire_data[1] & 0x80)
        assert wire_data[1] & ~0x80 == len(payload)
        if masked:
            maskbytes = itertools.cycle(bytearray(wire_data[2:6]))
            data = bytearray(b ^ next(maskbytes)
                             for b in bytearray(wire_data[6:]))
        else:
            data = wire_data[2:]
        assert data == payload

        # Verify that connection emits the PongReceive event with the correct
        # payload.
        me.receive_bytes(wire_data)
        event = next(me.events())
        assert isinstance(event, PongReceived)
        assert event.payload == payload
        with pytest.raises(StopIteration):
            repr(next(me.events()))

    @pytest.mark.parametrize('args, expected_payload', [
        ((), b''),
        ((b'abcdef',), b'abcdef')
    ], ids=['nopayload', 'payload'])
    def test_unsolicited_pong(self, args, expected_payload):
        client, server = self.create_connection()
        client.pong(*args)
        wire_data = client.bytes_to_send()
        server.receive_bytes(wire_data)
        events = list(server.events())
        assert len(events) == 1
        assert isinstance(events[0], PongReceived)
        assert events[0].payload == expected_payload

    @pytest.mark.parametrize('text,payload,full_message,full_frame', [
        (True, u'ƒñö®∂😎', True, True),
        (True, u'ƒñö®∂😎', False, True),
        (True, u'ƒñö®∂😎', False, False),
        (False, b'x' * 23, True, True),
        (False, b'x' * 23, False, True),
        (False, b'x' * 23, False, False),
    ])
    def test_data_events(self, text, payload, full_message, full_frame):
        if text:
            opcode = 0x01
            encoded_payload = payload.encode('utf8')
        else:
            opcode = 0x02
            encoded_payload = payload

        if full_message:
            opcode = bytearray([opcode | 0x80])
        else:
            opcode = bytearray([opcode])

        if full_frame:
            length = bytearray([len(encoded_payload)])
        else:
            length = bytearray([len(encoded_payload) + 100])

        frame = opcode + length + encoded_payload

        connection = WSConnection(CLIENT, host='localhost', resource='foo')
        connection._proto = FrameProtocol(True, [])
        connection._state = ConnectionState.OPEN
        connection.bytes_to_send()

        connection.receive_bytes(frame)
        event = next(connection.events())
        if text:
            assert isinstance(event, TextReceived)
        else:
            assert isinstance(event, BytesReceived)
        assert event.data == payload
        assert event.frame_finished is full_frame
        assert event.message_finished is full_message

        assert not connection.bytes_to_send()

    def test_frame_protocol_somehow_loses_its_mind(self):
        class FailFrame(object):
            opcode = object()

        class DoomProtocol(object):
            def receive_bytes(self, data):
                return None

            def received_frames(self):
                return [FailFrame()]

        connection = WSConnection(CLIENT, host='localhost', resource='foo')
        connection._proto = DoomProtocol()
        connection._state = ConnectionState.OPEN
        connection.bytes_to_send()

        connection.receive_bytes(b'')
        with pytest.raises(StopIteration):
            next(connection.events())
        assert not connection.bytes_to_send()

    def test_frame_protocol_gets_fed_garbage(self):
        client, server = self.create_connection()

        payload = b'x' * 23
        frame = b'\x09' + bytearray([len(payload)]) + payload

        client.receive_bytes(frame)
        event = next(client.events())
        assert isinstance(event, ConnectionClosed)
        assert event.code == CloseReason.PROTOCOL_ERROR

        output = client.bytes_to_send()
        assert output[:1] == b'\x88'



# -*- coding: utf-8 -*-
"""
wsproto/connection
~~~~~~~~~~~~~~~~~~

An implementation of a WebSocket connection.
"""

import os
import base64
import hashlib
from collections import deque

from enum import Enum

import h11

from .events import (
    ConnectionRequested, ConnectionEstablished, ConnectionClosed,
    ConnectionFailed, TextReceived, BytesReceived, PingReceived, PongReceived
)
from .frame_protocol import FrameProtocol, ParseFailed, CloseReason, Opcode
from .utilities import normed_header_dict, split_comma_header

# RFC6455, Section 1.3 - Opening Handshake
ACCEPT_GUID = b"258EAFA5-E914-47DA-95CA-C5AB0DC85B11"


class ConnectionState(Enum):
    """
    RFC 6455, Section 4 - Opening Handshake
    """
    CONNECTING = 0
    OPEN = 1
    CLOSING = 2
    CLOSED = 3

class ConnectionType(Enum):
    #: This connection will act as client and talk to a remote server
    CLIENT = 1

    #: This connection will as as server and waits for client connections
    SERVER = 2


CLIENT = ConnectionType.CLIENT
SERVER = ConnectionType.SERVER

class WSConnection(object):
    """
    A low-level WebSocket connection object.

    This wraps two other protocol objects, an HTTP/1.1 protocol object used
    to do the initial HTTP upgrade handshake and a WebSocket frame protocol
    object used to exchange messages and other control frames.

    :param conn_type: Whether this object is on the client- or server-side of
        a connection. To initialise as a client pass ``CLIENT`` otherwise
        pass ``SERVER``.
    :type conn_type: ``ConnectionType``

    :param host: The hostname to pass to the server when acting as a client.
    :type host: ``str``

    :param resource: The resource (aka path) to pass to the server when acting
        as a client.
    :type resource: ``str``

    :param extensions: A list of extensions to use on this connection.
        Defaults to to an empty list. Extensions should be instances of a
        subclass of :class:`Extension <wsproto.extensions.Extension>`.

    :param subprotocols: A list of subprotocols to request when acting as a
        client, ordered by preference. This has no impact on the connection
        itself. Defaults to an empty list.
    :type subprotocol: ``list`` of ``str``
    """

    def __init__(self, conn_type, host=None, resource=None, extensions=None,
                 subprotocols=None):
        self.client = conn_type is ConnectionType.CLIENT

        self.host = host
        self.resource = resource

        self.subprotocols = subprotocols or []
        self.extensions = extensions or []

        self.version = b'13'

        self._state = ConnectionState.CONNECTING
        self._close_reason = None

        self._nonce = None
        self._outgoing = b''
        self._events = deque()
        self._proto = None

        if self.client:
            self._upgrade_connection = h11.Connection(h11.CLIENT)
        else:
            self._upgrade_connection = h11.Connection(h11.SERVER)

        if self.client:
            if self.host is None:
                raise ValueError(
                    "Host must not be None for a client-side connection.")
            if self.resource is None:
                raise ValueError(
                    "Resource must not be None for a client-side connection.")
            self.initiate_connection()

    def initiate_connection(self):
        self._generate_nonce()

        headers = {
            b"Host": self.host.encode('ascii'),
            b"Upgrade": b'WebSocket',
            b"Connection": b'Upgrade',
            b"Sec-WebSocket-Key": self._nonce,
            b"Sec-WebSocket-Version": self.version,
        }

        if self.subprotocols:
            headers[b"Sec-WebSocket-Protocol"] = ", ".join(self.subprotocols)

        if self.extensions:
            offers = {e.name: e.offer(self) for e in self.extensions}
            extensions = []
            for name, params in offers.items():
                if params is True:
                    extensions.append(name.encode('ascii'))
                elif params:
                    # py34 annoyance: doesn't support bytestring formatting
                    extensions.append(('%s; %s' % (name, params))
                                      .encode("ascii"))
            if extensions:
                headers[b'Sec-WebSocket-Extensions'] = b', '.join(extensions)

        upgrade = h11.Request(method=b'GET', target=self.resource,
                              headers=headers.items())
        self._outgoing += self._upgrade_connection.send(upgrade)

    def send_data(self, payload, final=True):
        """
        Send a message or part of a message to the remote peer.

        If ``final`` is ``False`` it indicates that this is part of a longer
        message. If ``final`` is ``True`` it indicates that this is either a
        self-contained message or the last part of a longer message.

        If ``payload`` is of type ``bytes`` then the message is flagged as
        being binary. If it is of type ``str`` the message is encoded as UTF-8
        and sent as text.

        :param payload: The message body to send.
        :type payload: ``bytes`` or ``str``

        :param final: Whether there are more parts to this message to be sent.
        :type final: ``bool``
        """

        self._outgoing += self._proto.send_data(payload, final)

    def close(self, code=CloseReason.NORMAL_CLOSURE, reason=None):
        """
        Initiate the close handshake by sending a CLOSE control message.

        A clean teardown requires a CLOSE control messages from the other
        endpoint before the underlying TCP connection can be closed, see
        :class:`~wsproto.events.ConnectionClosed`.
        """
        self._outgoing += self._proto.close(code, reason)
        self._state = ConnectionState.CLOSING

    @property
    def closed(self):
        return self._state is ConnectionState.CLOSED

    def bytes_to_send(self, amount=None):
        """
        Returns some data for sending out of the internal data buffer.

        This method is analogous to ``read`` on a file-like object, but it
        doesn't block. Instead, it returns as much data as the user asks for,
        or less if that much data is not available. It does not perform any
        I/O, and so uses a different name.

        :param amount: (optional) The maximum amount of data to return. If not
            set, or set to ``None``, will return as much data as possible.
        :type amount: ``int``
        :returns: A bytestring containing the data to send on the wire.
        :rtype: ``bytes``
        """
        if amount is None:
            data = self._outgoing
            self._outgoing = b''
        else:
            data = self._outgoing[:amount]
            self._outgoing = self._outgoing[amount:]
        return data

    def receive_bytes(self, data):
        """
        Pass some received data to the connection for handling.

        A list of events that the remote peer triggered by sending this data can
        be retrieved with :meth:`~wsproto.connection.WSConnection.events`.

        :param data: The data received from the remote peer on the network.
        :type data: ``bytes``
        """

        if data is None and self._state is ConnectionState.OPEN:
            # "If _The WebSocket Connection is Closed_ and no Close control
            # frame was received by the endpoint (such as could occur if the
            # underlying transport connection is lost), _The WebSocket
            # Connection Close Code_ is considered to be 1006."
            self._events.append(ConnectionClosed(CloseReason.ABNORMAL_CLOSURE))
            self._state = ConnectionState.CLOSED
            return

        if self._state is ConnectionState.CONNECTING:
            event, data = self._process_upgrade(data)
            if event is not None:
                self._events.append(event)

        if self._state in (ConnectionState.OPEN, ConnectionState.CLOSING):
            self._proto.receive_bytes(data)
        elif self._state is ConnectionState.CLOSED:
            raise ValueError('Connection already closed.')

    def events(self):
        """
        Return a generator that provides any events that have been generated
        by protocol activity.

        :returns: generator of :class:`Event <wsproto.events.Event>` subclasses
        """

        while self._events:
            yield self._events.popleft()

        if self._proto is None:
            return

        try:
            for frame in self._proto.received_frames():
                if frame.opcode is Opcode.PING:
                    assert frame.frame_finished and frame.message_finished
                    self._outgoing += self._proto.pong(frame.payload)
                    yield PingReceived(frame.payload)

                elif frame.opcode is Opcode.PONG:
                    assert frame.frame_finished and frame.message_finished
                    yield PongReceived(frame.payload)

                elif frame.opcode is Opcode.CLOSE:
                    code, reason = frame.payload
                    if self._state is ConnectionState.OPEN:
                        self.close(code, reason)
                    self._state = ConnectionState.CLOSED
                    yield ConnectionClosed(code, reason)

                elif frame.opcode is Opcode.TEXT:
                    yield TextReceived(frame.payload,
                                       frame.frame_finished,
                                       frame.message_finished)

                elif frame.opcode is Opcode.BINARY:
                    yield BytesReceived(frame.payload,
                                        frame.frame_finished,
                                        frame.message_finished)
        except ParseFailed as exc:
            # XX FIXME: apparently autobahn intentionally deviates from the
            # spec in that on protocol errors it just closes the connection
            # rather than trying to send a CLOSE frame. Investigate whether we
            # should do the same.
            self.close(code=exc.code, reason=str(exc))
            yield ConnectionClosed(exc.code, reason=str(exc))

    def accept(self, event, subprotocol=None):
        request = event.h11request
        request_headers = normed_header_dict(request.headers)

        nonce = request_headers[b'sec-websocket-key']
        accept_token = self._generate_accept_token(nonce)

        headers = {
            b"Upgrade": b'WebSocket',
            b"Connection": b'Upgrade',
            b"Sec-WebSocket-Accept": accept_token,
        }

        if subprotocol is not None:
            if subprotocol not in event.proposed_subprotocols:
                raise ValueError(
                    "unexpected subprotocol {!r}".format(subprotocol))
            headers[b'Sec-WebSocket-Protocol'] = subprotocol

        extensions = request_headers.get(b'sec-websocket-extensions', None)
        if extensions:
            accepts = self._extension_accept(extensions)
            if accepts:
                headers[b"Sec-WebSocket-Extensions"] = accepts

        response = h11.InformationalResponse(status_code=101,
                                             headers=headers.items())
        self._outgoing += self._upgrade_connection.send(response)
        self._proto = FrameProtocol(self.client, self.extensions)
        self._state = ConnectionState.OPEN

    def ping(self, payload=None):
        """
        Send a PING message to the peer.

        :param payload: an optional payload to send with the message
        """

        payload = bytes(payload or b'')
        self._outgoing += self._proto.ping(payload)

    def pong(self, payload=None):
        """
        Send a PONG message to the peer.

        This method can be used to send an unsolicted PONG to the peer.
        It is not needed otherwise since every received PING causes a
        corresponding PONG to be sent automatically.

        :param payload: an optional payload to send with the message
        """

        payload = bytes(payload or b'')
        self._outgoing += self._proto.pong(payload)

    def _generate_nonce(self):
        # os.urandom may be overkill for this use case, but I don't think this
        # is a bottleneck, and better safe than sorry...
        self._nonce = base64.b64encode(os.urandom(16))

    def _generate_accept_token(self, token):
        accept_token = token + ACCEPT_GUID
        accept_token = hashlib.sha1(accept_token).digest()
        return base64.b64encode(accept_token)

    def _process_upgrade(self, data):
        self._upgrade_connection.receive_data(data)
        while True:
            try:
                event = self._upgrade_connection.next_event()
            except h11.RemoteProtocolError:
                return ConnectionFailed(CloseReason.PROTOCOL_ERROR,
                                        "Bad HTTP message"), b''
            if event is h11.NEED_DATA:
                break
            elif self.client and isinstance(event, (h11.InformationalResponse,
                                                    h11.Response)):
                data = self._upgrade_connection.trailing_data[0]
                return self._establish_client_connection(event), data
            elif not self.client and isinstance(event, h11.Request):
                return self._process_connection_request(event), None
            else:
                return ConnectionFailed(CloseReason.PROTOCOL_ERROR,
                                        "Bad HTTP message"), b''

        self._incoming = b''
        return None, None

    def _establish_client_connection(self, event):
        if event.status_code != 101:
            return ConnectionFailed(CloseReason.PROTOCOL_ERROR,
                                    "Bad status code from server")
        headers = normed_header_dict(event.headers)
        connection_tokens = split_comma_header(headers[b'connection'])
        if not any(token.lower() == 'upgrade' for token in connection_tokens):
            return ConnectionFailed(CloseReason.PROTOCOL_ERROR,
                                    "Missing Connection: Upgrade header")
        if headers[b'upgrade'].lower() != b'websocket':
            return ConnectionFailed(CloseReason.PROTOCOL_ERROR,
                                    "Missing Upgrade: WebSocket header")

        accept_token = self._generate_accept_token(self._nonce)
        if headers[b'sec-websocket-accept'] != accept_token:
            return ConnectionFailed(CloseReason.PROTOCOL_ERROR,
                                    "Bad accept token")

        subprotocol = headers.get(b'sec-websocket-protocol', None)
        if subprotocol is not None:
            subprotocol = subprotocol.decode('ascii')
            if subprotocol not in self.subprotocols:
                return ConnectionFailed(CloseReason.PROTOCOL_ERROR,
                                        "unrecognized subprotocol {!r}"
                                        .format(subprotocol))

        extensions = headers.get(b'sec-websocket-extensions', None)
        if extensions:
            accepts = split_comma_header(extensions)

            for accept in accepts:
                name = accept.split(';', 1)[0].strip()
                for extension in self.extensions:
                    if extension.name == name:
                        extension.finalize(self, accept)
                        break
                else:
                    return ConnectionFailed(CloseReason.PROTOCOL_ERROR,
                                            "unrecognized extension {!r}"
                                            .format(name))

        self._proto = FrameProtocol(self.client, self.extensions)
        self._state = ConnectionState.OPEN
        return ConnectionEstablished(subprotocol, extensions)

    def _process_connection_request(self, event):
        if event.method != b'GET':
            return ConnectionFailed(CloseReason.PROTOCOL_ERROR,
                                    "Request method must be GET")
        headers = normed_header_dict(event.headers)
        connection_tokens = split_comma_header(headers[b'connection'])
        if not any(token.lower() == 'upgrade' for token in connection_tokens):
            return ConnectionFailed(CloseReason.PROTOCOL_ERROR,
                                    "Missing Connection: Upgrade header")
        if headers[b'upgrade'].lower() != b'websocket':
            return ConnectionFailed(CloseReason.PROTOCOL_ERROR,
                                    "Missing Upgrade: WebSocket header")

        if b'sec-websocket-version' not in headers:
            return ConnectionFailed(CloseReason.PROTOCOL_ERROR,
                                    "Missing Sec-WebSocket-Version header")
        # XX FIXME: need to check Sec-Websocket-Version, and respond with a
        # 400 if it's not what we expect

        if b'sec-websocket-protocol' in headers:
            proposed_subprotocols = split_comma_header(
                headers[b'sec-websocket-protocol'])
        else:
            proposed_subprotocols = []

        if b'sec-websocket-key' not in headers:
            return ConnectionFailed(CloseReason.PROTOCOL_ERROR,
                                    "Missing Sec-WebSocket-Key header")

        return ConnectionRequested(proposed_subprotocols, event)

    def _extension_accept(self, extensions_header):
        accepts = {}
        offers = split_comma_header(extensions_header)

        for offer in offers:
            name = offer.split(';', 1)[0].strip()
            for extension in self.extensions:
                if extension.name == name:
                    accept = extension.accept(self, offer)
                    if accept is True:
                        accepts[extension.name] = True
                    elif accept is not False and accept is not None:
                        accepts[extension.name] = accept.encode('ascii')

        if accepts:
            extensions = []
            for name, params in accepts.items():
                if params is True:
                    extensions.append(name.encode('ascii'))
                else:
                    # py34 annoyance: doesn't support bytestring formatting
                    params = params.decode("ascii")
                    if params == '':
                        extensions.append(('%s' % (name))
                                          .encode("ascii"))
                    else:
                        extensions.append(('%s; %s' % (name, params))
                                          .encode("ascii"))
            return b', '.join(extensions)

        return None



# -*- coding: utf-8 -*-
"""
wsproto/extensions
~~~~~~~~~~~~~~~~~~

WebSocket extensions.
"""

import zlib

from .frame_protocol import CloseReason, Opcode, RsvBits


class Extension(object):
    name = None

    def enabled(self):
        return False

    def offer(self, connection):
        pass

    def accept(self, connection, offer):
        pass

    def finalize(self, connection, offer):
        pass

    def frame_inbound_header(self, proto, opcode, rsv, payload_length):
        return RsvBits(False, False, False)

    def frame_inbound_payload_data(self, proto, data):
        return data

    def frame_inbound_complete(self, proto, fin):
        pass

    def frame_outbound(self, proto, opcode, rsv, data, fin):
        return (rsv, data)


class PerMessageDeflate(Extension):
    name = 'permessage-deflate'

    DEFAULT_CLIENT_MAX_WINDOW_BITS = 15
    DEFAULT_SERVER_MAX_WINDOW_BITS = 15

    def __init__(self, client_no_context_takeover=False,
                 client_max_window_bits=None, server_no_context_takeover=False,
                 server_max_window_bits=None):
        self.client_no_context_takeover = client_no_context_takeover
        if client_max_window_bits is None:
            client_max_window_bits = self.DEFAULT_CLIENT_MAX_WINDOW_BITS
        self.client_max_window_bits = client_max_window_bits
        self.server_no_context_takeover = server_no_context_takeover
        if server_max_window_bits is None:
            server_max_window_bits = self.DEFAULT_SERVER_MAX_WINDOW_BITS
        self.server_max_window_bits = server_max_window_bits

        self._compressor = None
        self._decompressor = None
        # This refers to the current frame
        self._inbound_is_compressible = None
        # This refers to the ongoing message (which might span multiple
        # frames). Only the first frame in a fragmented message is flagged for
        # compression, so this carries that bit forward.
        self._inbound_compressed = None

        self._enabled = False

    def _compressible_opcode(self, opcode):
        return opcode in (Opcode.TEXT, Opcode.BINARY, Opcode.CONTINUATION)

    def enabled(self):
        return self._enabled

    def offer(self, connection):
        parameters = [
            'client_max_window_bits=%d' % self.client_max_window_bits,
            'server_max_window_bits=%d' % self.server_max_window_bits,
            ]

        if self.client_no_context_takeover:
            parameters.append('client_no_context_takeover')
        if self.server_no_context_takeover:
            parameters.append('server_no_context_takeover')

        return '; '.join(parameters)

    def finalize(self, connection, offer):
        bits = [b.strip() for b in offer.split(';')]
        for bit in bits[1:]:
            if bit.startswith('client_no_context_takeover'):
                self.client_no_context_takeover = True
            elif bit.startswith('server_no_context_takeover'):
                self.server_no_context_takeover = True
            elif bit.startswith('client_max_window_bits'):
                self.client_max_window_bits = int(bit.split('=', 1)[1].strip())
            elif bit.startswith('server_max_window_bits'):
                self.server_max_window_bits = int(bit.split('=', 1)[1].strip())

        self._enabled = True

    def _parse_params(self, params):
        client_max_window_bits = None
        server_max_window_bits = None

        bits = [b.strip() for b in params.split(';')]
        for bit in bits[1:]:
            if bit.startswith('client_no_context_takeover'):
                self.client_no_context_takeover = True
            elif bit.startswith('server_no_context_takeover'):
                self.server_no_context_takeover = True
            elif bit.startswith('client_max_window_bits'):
                if '=' in bit:
                    client_max_window_bits = int(bit.split('=', 1)[1].strip())
                else:
                    client_max_window_bits = self.client_max_window_bits
            elif bit.startswith('server_max_window_bits'):
                if '=' in bit:
                    server_max_window_bits = int(bit.split('=', 1)[1].strip())
                else:
                    server_max_window_bits = self.server_max_window_bits

        return client_max_window_bits, server_max_window_bits

    def accept(self, connection, offer):
        client_max_window_bits, server_max_window_bits = \
            self._parse_params(offer)

        self._enabled = True

        parameters = []

        if self.client_no_context_takeover:
            parameters.append('client_no_context_takeover')
        if client_max_window_bits is not None:
            parameters.append('client_max_window_bits=%d' %
                              client_max_window_bits)
            self.client_max_window_bits = client_max_window_bits
        if self.server_no_context_takeover:
            parameters.append('server_no_context_takeover')
        if server_max_window_bits is not None:
            parameters.append('server_max_window_bits=%d' %
                              server_max_window_bits)
            self.server_max_window_bits = server_max_window_bits

        return '; '.join(parameters)

    def frame_inbound_header(self, proto, opcode, rsv, payload_length):
        if rsv.rsv1 and opcode.iscontrol():
            return CloseReason.PROTOCOL_ERROR
        if rsv.rsv1 and opcode is Opcode.CONTINUATION:
            return CloseReason.PROTOCOL_ERROR

        self._inbound_is_compressible = self._compressible_opcode(opcode)

        if self._inbound_compressed is None:
            self._inbound_compressed = rsv.rsv1
            if self._inbound_compressed:
                assert self._inbound_is_compressible
                if proto.client:
                    bits = self.server_max_window_bits
                else:
                    bits = self.client_max_window_bits
                if self._decompressor is None:
                    self._decompressor = zlib.decompressobj(-int(bits))

        return RsvBits(True, False, False)

    def frame_inbound_payload_data(self, proto, data):
        if not self._inbound_compressed or not self._inbound_is_compressible:
            return data

        try:
            return self._decompressor.decompress(bytes(data))
        except zlib.error:
            return CloseReason.INVALID_FRAME_PAYLOAD_DATA

    def frame_inbound_complete(self, proto, fin):
        if not fin:
            return None
        if not self._inbound_is_compressible:
            self._inbound_compressed = None
            return None
        if not self._inbound_compressed:
            self._inbound_compressed = None
            return None

        try:
            data = self._decompressor.decompress(b'\x00\x00\xff\xff')
            data += self._decompressor.flush()
        except zlib.error:
            return CloseReason.INVALID_FRAME_PAYLOAD_DATA

        if proto.client:
            no_context_takeover = self.server_no_context_takeover
        else:
            no_context_takeover = self.client_no_context_takeover

        if no_context_takeover:
            self._decompressor = None

        self._inbound_compressed = None

        return data

    def frame_outbound(self, proto, opcode, rsv, data, fin):
        if not self._compressible_opcode(opcode):
            return (rsv, data)

        if opcode is not Opcode.CONTINUATION:
            rsv = RsvBits(True, *rsv[1:])

        if self._compressor is None:
            assert opcode is not Opcode.CONTINUATION
            if proto.client:
                bits = self.client_max_window_bits
            else:
                bits = self.server_max_window_bits
            self._compressor = zlib.compressobj(zlib.Z_DEFAULT_COMPRESSION,
                                                zlib.DEFLATED, -int(bits))

        data = self._compressor.compress(bytes(data))

        if fin:
            data += self._compressor.flush(zlib.Z_SYNC_FLUSH)
            data = data[:-4]

            if proto.client:
                no_context_takeover = self.client_no_context_takeover
            else:
                no_context_takeover = self.server_no_context_takeover

            if no_context_takeover:
                self._compressor = None

        return (rsv, data)

    def __repr__(self):
        descr = ['client_max_window_bits=%d' % self.client_max_window_bits]
        if self.client_no_context_takeover:
            descr.append('client_no_context_takeover')
        descr.append('server_max_window_bits=%d' % self.server_max_window_bits)
        if self.server_no_context_takeover:
            descr.append('server_no_context_takeover')

        descr = '; '.join(descr)

        return '<%s %s>' % (self.__class__.__name__, descr)


#: SUPPORTED_EXTENSIONS maps all supported extension names to their class.
#: This can be used to iterate all supported extensions of wsproto, instantiate
#: new extensions based on their name, or check if a given extension is
#: supported or not.
SUPPORTED_EXTENSIONS = {
    PerMessageDeflate.name: PerMessageDeflate
}



import json
import socket

from wsproto.compat import PY2
from wsproto.connection import WSConnection, CLIENT, ConnectionEstablished, \
                               ConnectionClosed
from wsproto.events import TextReceived, DataReceived
from wsproto.extensions import PerMessageDeflate

if PY2:
    from urlparse import urlparse
else:
    from urllib.parse import urlparse

SERVER = 'ws://127.0.0.1:8642'
AGENT = 'wsproto'

if PY2:
    CONNECTION_EXCEPTIONS = (OSError,)
else:
    CONNECTION_EXCEPTIONS = (ConnectionError, OSError)

def get_case_count(server):
    uri = urlparse(server + '/getCaseCount')
    connection = WSConnection(CLIENT, uri.netloc, uri.path)
    sock = socket.socket()
    sock.connect((uri.hostname, uri.port or 80))

    sock.sendall(connection.bytes_to_send())

    case_count = None
    while case_count is None:
        data = sock.recv(65535)
        connection.receive_bytes(data)
        data = ""
        for event in connection.events():
            if isinstance(event, TextReceived):
                data += event.data
                if event.message_finished:
                    case_count = json.loads(data)
                    connection.close()
            try:
                sock.sendall(connection.bytes_to_send())
            except CONNECTION_EXCEPTIONS:
                break

    sock.close()
    return case_count

def run_case(server, case, agent):
    uri = urlparse(server + '/runCase?case=%d&agent=%s' % (case, agent))
    connection = WSConnection(CLIENT,
                              uri.netloc, '%s?%s' % (uri.path, uri.query),
                              extensions=[PerMessageDeflate()])
    sock = socket.socket()
    sock.connect((uri.hostname, uri.port or 80))

    sock.sendall(connection.bytes_to_send())
    closed = False

    while not closed:
        try:
            data = sock.recv(65535)
        except CONNECTION_EXCEPTIONS:
            data = None
        connection.receive_bytes(data or None)
        for event in connection.events():
            if isinstance(event, DataReceived):
                connection.send_data(event.data, event.message_finished)
            elif isinstance(event, ConnectionClosed):
                closed = True
            # else:
            #     print("??", event)
        if data is None:
            break
        try:
            data = connection.bytes_to_send()
            sock.sendall(data)
        except CONNECTION_EXCEPTIONS:
            closed = True
            break

def update_reports(server, agent):
    uri = urlparse(server + '/updateReports?agent=%s' % agent)
    connection = WSConnection(CLIENT,
                              uri.netloc, '%s?%s' % (uri.path, uri.query))
    sock = socket.socket()
    sock.connect((uri.hostname, uri.port or 80))

    sock.sendall(connection.bytes_to_send())
    closed = False

    while not closed:
        data = sock.recv(65535)
        connection.receive_bytes(data)
        for event in connection.events():
            if isinstance(event, ConnectionEstablished):
                connection.close()
                sock.sendall(connection.bytes_to_send())
                try:
                    sock.close()
                except CONNECTION_EXCEPTIONS:
                    pass
                finally:
                    closed = True

CASE = None
# 1.1.1 = 1
# 2.1 = 17
# 3.1 = 28
# 4.1.1 = 34
# 5.1 = 44
# 6.1.1 = 64
# 12.1.1 = 304
# 13.1.1 = 394

def run_tests(server, agent):
    case_count = get_case_count(server)
    if CASE is not None:
        print(">>>>> Running test case %d" % CASE)
        run_case(server, CASE, agent)
    else:
        for case in range(1, case_count + 1):
            print(">>>>> Running test case %d of %d" % (case, case_count))
            run_case(server, case, agent)
        print("\nRan %d cases." % case_count)
    update_reports(server, agent)

if __name__ == '__main__':
    run_tests(SERVER, AGENT)



# -*- coding: utf-8 -*-
"""
Test the HTTP upgrade phase of connection
"""

import base64
import email
import random

import pytest

from wsproto.compat import PY3
from wsproto.connection import WSConnection, CLIENT, SERVER
from wsproto.events import (
    ConnectionEstablished, ConnectionFailed, ConnectionRequested
)
from wsproto.extensions import Extension


def parse_headers(headers):
    if PY3:
        headers = email.message_from_bytes(headers)
    else:
        headers = email.message_from_string(headers)

    return dict(headers.items())


class FakeExtension(Extension):
    name = 'fake'

    def __init__(self, offer_response=None, accept_response=None):
        self.offer_response = offer_response
        self.accepted_offer = None
        self.offered = None
        self.accept_response = accept_response

    def offer(self, proto):
        return self.offer_response

    def finalize(self, proto, offer):
        self.accepted_offer = offer

    def accept(self, proto, offer):
        self.offered = offer
        return self.accept_response


class TestClientUpgrade(object):
    def initiate(self, host, path, **kwargs):
        ws = WSConnection(CLIENT, host, path, **kwargs)

        data = ws.bytes_to_send()
        request, headers = data.split(b'\r\n', 1)
        method, path, version = request.strip().split()
        headers = parse_headers(headers)

        return ws, method, path, version, headers

    def test_initiate_connection(self):
        _host = 'frob.nitz'
        _path = '/fnord'

        ws, method, path, version, headers = self.initiate(
            _host, _path, subprotocols=["foo", "bar"])

        assert method == b'GET'
        assert path == _path.encode('ascii')

        assert headers['host'] == _host
        assert headers['connection'].lower() == 'upgrade'
        assert headers['upgrade'].lower() == 'websocket'
        assert 'sec-websocket-key' in headers
        assert 'sec-websocket-version' in headers
        assert headers['sec-websocket-protocol'] == 'foo, bar'

    def test_no_subprotocols(self):
        ws, method, path, version, headers = self.initiate("foo", "/bar")
        assert 'sec-websocket-protocol' not in headers

    def test_correct_accept_token(self):
        _host = 'frob.nitz'
        _path = '/fnord'

        ws, method, path, version, headers = self.initiate(_host, _path)

        key = headers['sec-websocket-key'].encode('ascii')
        accept_token = ws._generate_accept_token(key)

        response = b"HTTP/1.1 101 Switching Protocols\r\n"
        response += b"Connection: Upgrade\r\n"
        response += b"Upgrade: WebSocket\r\n"
        response += b"Sec-WebSocket-Accept: " + accept_token + b"\r\n"
        response += b"\r\n"

        ws.receive_bytes(response)
        assert isinstance(next(ws.events()), ConnectionEstablished)

    def test_incorrect_accept_token(self):
        _host = 'frob.nitz'
        _path = '/fnord'

        ws, method, path, version, headers = self.initiate(_host, _path)

        key = b'This is wrong token'
        accept_token = ws._generate_accept_token(key)

        response = b"HTTP/1.1 101 Switching Protocols\r\n"
        response += b"Connection: Upgrade\r\n"
        response += b"Upgrade: WebSocket\r\n"
        response += b"Sec-WebSocket-Accept: " + accept_token + b"\r\n"
        response += b"\r\n"

        ws.receive_bytes(response)
        assert isinstance(next(ws.events()), ConnectionFailed)

    def test_bad_connection_header(self):
        _host = 'frob.nitz'
        _path = '/fnord'

        ws, method, path, version, headers = self.initiate(_host, _path)

        key = headers['sec-websocket-key'].encode('ascii')
        accept_token = ws._generate_accept_token(key)

        response = b"HTTP/1.1 101 Switching Protocols\r\n"
        response += b"Connection: Updraft\r\n"
        response += b"Upgrade: WebSocket\r\n"
        response += b"Sec-WebSocket-Accept: " + accept_token + b"\r\n"
        response += b"\r\n"

        ws.receive_bytes(response)
        assert isinstance(next(ws.events()), ConnectionFailed)

    def test_bad_upgrade_header(self):
        _host = 'frob.nitz'
        _path = '/fnord'

        ws, method, path, version, headers = self.initiate(_host, _path)

        key = headers['sec-websocket-key'].encode('ascii')
        accept_token = ws._generate_accept_token(key)

        response = b"HTTP/1.1 101 Switching Protocols\r\n"
        response += b"Connection: Upgrade\r\n"
        response += b"Upgrade: SebWocket\r\n"
        response += b"Sec-WebSocket-Accept: " + accept_token + b"\r\n"
        response += b"\r\n"

        ws.receive_bytes(response)
        assert isinstance(next(ws.events()), ConnectionFailed)

    def test_simple_extension_offer(self):
        _host = 'frob.nitz'
        _path = '/fnord'
        _ext = FakeExtension(offer_response=True)

        ws, method, path, version, headers = \
            self.initiate(_host, _path, extensions=[_ext])

        assert _ext.name == headers['sec-websocket-extensions']

    def test_simple_extension_non_offer(self):
        _host = 'frob.nitz'
        _path = '/fnord'
        _ext = FakeExtension(offer_response=False)

        ws, method, path, version, headers = \
            self.initiate(_host, _path, extensions=[_ext])

        assert 'sec-websocket-extensions' not in headers

    def test_extension_offer_with_params(self):
        ext_parameters = 'parameter1=value1; parameter2=value2'
        _ext = FakeExtension(offer_response=ext_parameters)

        _host = 'frob.nitz'
        _path = '/fnord'

        ws, method, path, version, headers = \
            self.initiate(_host, _path, extensions=[_ext])

        assert headers['sec-websocket-extensions'] == \
            '%s; %s' % (_ext.name, ext_parameters)

    def test_simple_extension_accept(self):
        _host = 'frob.nitz'
        _path = '/fnord'
        _ext = FakeExtension(offer_response=True)

        ws, method, path, version, headers = \
            self.initiate(_host, _path, extensions=[_ext])

        key = headers['sec-websocket-key'].encode('ascii')
        accept_token = ws._generate_accept_token(key)

        response = b"HTTP/1.1 101 Switching Protocols\r\n"
        response += b"Connection: Upgrade\r\n"
        response += b"Upgrade: WebSocket\r\n"
        response += b"Sec-WebSocket-Accept: " + accept_token + b"\r\n"
        response += b"Sec-WebSocket-Extensions: " + \
                    _ext.name.encode('ascii') + b"\r\n"
        response += b"\r\n"

        ws.receive_bytes(response)
        assert isinstance(next(ws.events()), ConnectionEstablished)
        assert _ext.name in _ext.accepted_offer

    def test_extension_accept_with_parameters(self):
        _host = 'frob.nitz'
        _path = '/fnord'
        _ext = FakeExtension(offer_response=True)

        ws, method, path, version, headers = \
            self.initiate(_host, _path, extensions=[_ext])

        key = headers['sec-websocket-key'].encode('ascii')
        accept_token = ws._generate_accept_token(key)
        ext_parameters = 'parameter1=value1; parameter2=value2'
        extensions = _ext.name + '; ' + ext_parameters

        response = b"HTTP/1.1 101 Switching Protocols\r\n"
        response += b"Connection: Upgrade\r\n"
        response += b"Upgrade: WebSocket\r\n"
        response += b"Sec-WebSocket-Accept: " + accept_token + b"\r\n"
        response += b"Sec-WebSocket-Extensions: " + \
                    extensions.encode('ascii') + b"\r\n"
        response += b"\r\n"

        ws.receive_bytes(response)
        assert isinstance(next(ws.events()), ConnectionEstablished)
        assert _ext.accepted_offer == extensions

    def test_accept_an_extension_we_do_not_recognise(self):
        _host = 'frob.nitz'
        _path = '/fnord'
        _ext = FakeExtension(offer_response=True)

        ws, method, path, version, headers = \
            self.initiate(_host, _path, extensions=[_ext])

        key = headers['sec-websocket-key'].encode('ascii')
        accept_token = ws._generate_accept_token(key)

        response = b"HTTP/1.1 101 Switching Protocols\r\n"
        response += b"Connection: Upgrade\r\n"
        response += b"Upgrade: WebSocket\r\n"
        response += b"Sec-WebSocket-Accept: " + accept_token + b"\r\n"
        response += b"Sec-WebSocket-Extensions: pretend\r\n"
        response += b"\r\n"

        ws.receive_bytes(response)
        assert isinstance(next(ws.events()), ConnectionFailed)

    def test_wrong_status_code_in_response(self):
        _host = 'frob.nitz'
        _path = '/fnord'

        ws, method, path, version, headers = self.initiate(_host, _path)

        response = b"HTTP/1.1 200 OK\r\n"
        response += b"Server: SimpleHTTP/0.6 Python/3.6.1\r\n"
        response += b"Date: Fri, 02 Jun 2017 20:40:39 GMT\r\n"
        response += b"Content-type: application/octet-stream\r\n"
        response += b"Content-Length: 0\r\n"
        response += b"Last-Modified: Fri, 02 Jun 2017 20:40:00 GMT\r\n"
        response += b"Connection: close\r\n"
        response += b"\r\n"

        ws.receive_bytes(response)
        assert isinstance(next(ws.events()), ConnectionFailed)

    def test_response_takes_a_few_goes(self):
        _host = 'frob.nitz'
        _path = '/fnord'

        ws, method, path, version, headers = self.initiate(_host, _path)

        key = headers['sec-websocket-key'].encode('ascii')
        accept_token = ws._generate_accept_token(key)

        response = b"HTTP/1.1 101 Switching Protocols\r\n"
        response += b"Connection: Upgrade\r\n"
        response += b"Upgrade: WebSocket\r\n"
        response += b"Sec-WebSocket-Accept: " + accept_token + b"\r\n"
        response += b"\r\n"

        split = len(response) // 2

        ws.receive_bytes(response[:split])
        with pytest.raises(StopIteration):
            next(ws.events())

        ws.receive_bytes(response[split:])
        assert isinstance(next(ws.events()), ConnectionEstablished)

    def test_subprotocol_offer(self):
        _host = 'frob.nitz'
        _path = '/fnord'
        subprotocols = ['one', 'two']

        ws, method, path, version, headers = \
            self.initiate(_host, _path, subprotocols=subprotocols)

        for subprotocol in subprotocols:
            assert subprotocol in headers['sec-websocket-protocol']

    def test_subprotocol_accept(self):
        _host = 'frob.nitz'
        _path = '/fnord'
        subprotocols = ['one', 'two']

        ws, method, path, version, headers = \
            self.initiate(_host, _path, subprotocols=subprotocols)

        key = headers['sec-websocket-key'].encode('ascii')
        accept_token = ws._generate_accept_token(key)

        response = b"HTTP/1.1 101 Switching Protocols\r\n"
        response += b"Connection: Upgrade\r\n"
        response += b"Upgrade: WebSocket\r\n"
        response += b"Sec-WebSocket-Accept: " + accept_token + b"\r\n"
        response += b"Sec-WebSocket-Protocol: " + \
                    subprotocols[0].encode('ascii') + b"\r\n"
        response += b"\r\n"

        ws.receive_bytes(response)
        event = next(ws.events())
        assert isinstance(event, ConnectionEstablished)
        assert event.subprotocol == subprotocols[0]

    def test_subprotocol_accept_unoffered(self):
        _host = 'frob.nitz'
        _path = '/fnord'
        subprotocols = ['one', 'two']

        ws, method, path, version, headers = \
            self.initiate(_host, _path, subprotocols=subprotocols)

        key = headers['sec-websocket-key'].encode('ascii')
        accept_token = ws._generate_accept_token(key)

        response = b"HTTP/1.1 101 Switching Protocols\r\n"
        response += b"Connection: Upgrade\r\n"
        response += b"Upgrade: WebSocket\r\n"
        response += b"Sec-WebSocket-Accept: " + accept_token + b"\r\n"
        response += b"Sec-WebSocket-Protocol: three\r\n"
        response += b"\r\n"

        ws.receive_bytes(response)
        assert isinstance(next(ws.events()), ConnectionFailed)


class TestServerUpgrade(object):
    def test_correct_request(self):
        test_host = 'frob.nitz'
        test_path = '/fnord'

        ws = WSConnection(SERVER)

        nonce = bytes(random.getrandbits(8) for x in range(0, 16))
        nonce = base64.b64encode(nonce)

        request = b"GET " + test_path.encode('ascii') + b" HTTP/1.1\r\n"
        request += b'Host: ' + test_host.encode('ascii') + b'\r\n'
        request += b'Connection: Upgrade\r\n'
        request += b'Upgrade: WebSocket\r\n'
        request += b'Sec-WebSocket-Version: 13\r\n'
        request += b'Sec-WebSocket-Key: ' + nonce + b'\r\n'
        request += b'\r\n'

        ws.receive_bytes(request)
        event = next(ws.events())
        assert isinstance(event, ConnectionRequested)
        ws.accept(event)

        data = ws.bytes_to_send()
        response, headers = data.split(b'\r\n', 1)
        version, code, reason = response.split(b' ')
        headers = parse_headers(headers)

        accept_token = ws._generate_accept_token(nonce)

        assert int(code) == 101
        assert headers['connection'].lower() == 'upgrade'
        assert headers['upgrade'].lower() == 'websocket'
        assert headers['sec-websocket-accept'] == accept_token.decode('ascii')

    def test_correct_request_expanded_connection_header(self):
        test_host = 'frob.nitz'
        test_path = '/fnord'

        ws = WSConnection(SERVER)

        nonce = bytes(random.getrandbits(8) for x in range(0, 16))
        nonce = base64.b64encode(nonce)

        request = b"GET " + test_path.encode('ascii') + b" HTTP/1.1\r\n"
        request += b'Host: ' + test_host.encode('ascii') + b'\r\n'
        request += b'Connection: keep-alive, Upgrade\r\n'
        request += b'Upgrade: WebSocket\r\n'
        request += b'Sec-WebSocket-Version: 13\r\n'
        request += b'Sec-WebSocket-Key: ' + nonce + b'\r\n'
        request += b'\r\n'

        ws.receive_bytes(request)
        event = next(ws.events())
        assert isinstance(event, ConnectionRequested)

    def test_wrong_method(self):
        test_host = 'frob.nitz'
        test_path = '/fnord'

        ws = WSConnection(SERVER)

        nonce = bytes(random.getrandbits(8) for x in range(0, 16))
        nonce = base64.b64encode(nonce)

        request = b'POST ' + test_path.encode('ascii') + b' HTTP/1.1\r\n'
        request += b'Host: ' + test_host.encode('ascii') + b'\r\n'
        request += b'Connection: Upgrade\r\n'
        request += b'Upgrade: WebSocket\r\n'
        request += b'Sec-WebSocket-Version: 13\r\n'
        request += b'Sec-WebSocket-Key: ' + nonce + b'\r\n'
        request += b'\r\n'

        ws.receive_bytes(request)
        event = next(ws.events())
        assert isinstance(event, ConnectionFailed)

    def test_bad_connection(self):
        test_host = 'frob.nitz'
        test_path = '/fnord'

        ws = WSConnection(SERVER)

        nonce = bytes(random.getrandbits(8) for x in range(0, 16))
        nonce = base64.b64encode(nonce)

        request = b'GET ' + test_path.encode('ascii') + b' HTTP/1.1\r\n'
        request += b'Host: ' + test_host.encode('ascii') + b'\r\n'
        request += b'Connection: Zoinks\r\n'
        request += b'Upgrade: WebSocket\r\n'
        request += b'Sec-WebSocket-Version: 13\r\n'
        request += b'Sec-WebSocket-Key: ' + nonce + b'\r\n'
        request += b'\r\n'

        ws.receive_bytes(request)
        event = next(ws.events())
        assert isinstance(event, ConnectionFailed)

    def test_bad_upgrade(self):
        test_host = 'frob.nitz'
        test_path = '/fnord'

        ws = WSConnection(SERVER)

        nonce = bytes(random.getrandbits(8) for x in range(0, 16))
        nonce = base64.b64encode(nonce)

        request = b'GET ' + test_path.encode('ascii') + b' HTTP/1.1\r\n'
        request += b'Host: ' + test_host.encode('ascii') + b'\r\n'
        request += b'Connection: Upgrade\r\n'
        request += b'Upgrade: WebPocket\r\n'
        request += b'Sec-WebSocket-Version: 13\r\n'
        request += b'Sec-WebSocket-Key: ' + nonce + b'\r\n'
        request += b'\r\n'

        ws.receive_bytes(request)
        event = next(ws.events())
        assert isinstance(event, ConnectionFailed)

    def test_missing_version(self):
        test_host = 'frob.nitz'
        test_path = '/fnord'

        ws = WSConnection(SERVER)

        nonce = bytes(random.getrandbits(8) for x in range(0, 16))
        nonce = base64.b64encode(nonce)

        request = b'GET ' + test_path.encode('ascii') + b' HTTP/1.1\r\n'
        request += b'Host: ' + test_host.encode('ascii') + b'\r\n'
        request += b'Connection: Upgrade\r\n'
        request += b'Upgrade: WebSocket\r\n'
        request += b'Sec-WebSocket-Key: ' + nonce + b'\r\n'
        request += b'\r\n'

        ws.receive_bytes(request)
        event = next(ws.events())
        assert isinstance(event, ConnectionFailed)

    def test_missing_key(self):
        test_host = 'frob.nitz'
        test_path = '/fnord'

        ws = WSConnection(SERVER)

        request = b'GET ' + test_path.encode('ascii') + b' HTTP/1.1\r\n'
        request += b'Host: ' + test_host.encode('ascii') + b'\r\n'
        request += b'Connection: Upgrade\r\n'
        request += b'Upgrade: WebSocket\r\n'
        request += b'Sec-WebSocket-Version: 13\r\n'
        request += b'\r\n'

        ws.receive_bytes(request)
        event = next(ws.events())
        assert isinstance(event, ConnectionFailed)

    def test_subprotocol_offers(self):
        test_host = 'frob.nitz'
        test_path = '/fnord'

        ws = WSConnection(SERVER)

        nonce = bytes(random.getrandbits(8) for x in range(0, 16))
        nonce = base64.b64encode(nonce)

        request = b'GET ' + test_path.encode('ascii') + b' HTTP/1.1\r\n'
        request += b'Host: ' + test_host.encode('ascii') + b'\r\n'
        request += b'Connection: Upgrade\r\n'
        request += b'Upgrade: WebSocket\r\n'
        request += b'Sec-WebSocket-Version: 13\r\n'
        request += b'Sec-WebSocket-Key: ' + nonce + b'\r\n'
        request += b'Sec-WebSocket-Protocol: one, two\r\n'
        request += b'\r\n'

        ws.receive_bytes(request)
        event = next(ws.events())
        assert isinstance(event, ConnectionRequested)
        assert event.proposed_subprotocols == ['one', 'two']

    def test_accept_subprotocol(self):
        test_host = 'frob.nitz'
        test_path = '/fnord'

        ws = WSConnection(SERVER)

        nonce = bytes(random.getrandbits(8) for x in range(0, 16))
        nonce = base64.b64encode(nonce)

        request = b'GET ' + test_path.encode('ascii') + b' HTTP/1.1\r\n'
        request += b'Host: ' + test_host.encode('ascii') + b'\r\n'
        request += b'Connection: Upgrade\r\n'
        request += b'Upgrade: WebSocket\r\n'
        request += b'Sec-WebSocket-Version: 13\r\n'
        request += b'Sec-WebSocket-Key: ' + nonce + b'\r\n'
        request += b'Sec-WebSocket-Protocol: one, two\r\n'
        request += b'\r\n'

        ws.receive_bytes(request)
        event = next(ws.events())
        assert isinstance(event, ConnectionRequested)
        assert event.proposed_subprotocols == ['one', 'two']

        ws.accept(event, 'two')

        data = ws.bytes_to_send()
        response, headers = data.split(b'\r\n', 1)
        version, code, reason = response.split(b' ')
        headers = parse_headers(headers)

        assert int(code) == 101
        assert headers['sec-websocket-protocol'] == 'two'

    def test_accept_wrong_subprotocol(self):
        test_host = 'frob.nitz'
        test_path = '/fnord'

        ws = WSConnection(SERVER)

        nonce = bytes(random.getrandbits(8) for x in range(0, 16))
        nonce = base64.b64encode(nonce)

        request = b'GET ' + test_path.encode('ascii') + b' HTTP/1.1\r\n'
        request += b'Host: ' + test_host.encode('ascii') + b'\r\n'
        request += b'Connection: Upgrade\r\n'
        request += b'Upgrade: WebSocket\r\n'
        request += b'Sec-WebSocket-Version: 13\r\n'
        request += b'Sec-WebSocket-Key: ' + nonce + b'\r\n'
        request += b'Sec-WebSocket-Protocol: one, two\r\n'
        request += b'\r\n'

        ws.receive_bytes(request)
        event = next(ws.events())
        assert isinstance(event, ConnectionRequested)
        assert event.proposed_subprotocols == ['one', 'two']

        with pytest.raises(ValueError):
            ws.accept(event, 'three')

    def test_simple_extension_negotiation(self):
        test_host = 'frob.nitz'
        test_path = '/fnord'
        ext = FakeExtension(accept_response=True)

        ws = WSConnection(SERVER, extensions=[ext])

        nonce = bytes(random.getrandbits(8) for x in range(0, 16))
        nonce = base64.b64encode(nonce)

        request = b"GET " + test_path.encode('ascii') + b" HTTP/1.1\r\n"
        request += b'Host: ' + test_host.encode('ascii') + b'\r\n'
        request += b'Connection: Upgrade\r\n'
        request += b'Upgrade: WebSocket\r\n'
        request += b'Sec-WebSocket-Version: 13\r\n'
        request += b'Sec-WebSocket-Key: ' + nonce + b'\r\n'
        request += b'Sec-WebSocket-Extensions: ' + \
            ext.name.encode('ascii') + b'\r\n'
        request += b'\r\n'

        ws.receive_bytes(request)
        event = next(ws.events())
        assert isinstance(event, ConnectionRequested)
        ws.accept(event)

        data = ws.bytes_to_send()
        response, headers = data.split(b'\r\n', 1)
        version, code, reason = response.split(b' ')
        headers = parse_headers(headers)

        assert ext.offered == ext.name
        assert headers['sec-websocket-extensions'] == ext.name

    def test_extension_negotiation_with_our_parameters(self):
        test_host = 'frob.nitz'
        test_path = '/fnord'
        offered_params = 'parameter1=value3; parameter2=value4'
        ext_params = 'parameter1=value1; parameter2=value2'
        ext = FakeExtension(accept_response=ext_params)

        ws = WSConnection(SERVER, extensions=[ext])

        nonce = bytes(random.getrandbits(8) for x in range(0, 16))
        nonce = base64.b64encode(nonce)

        request = b"GET " + test_path.encode('ascii') + b" HTTP/1.1\r\n"
        request += b'Host: ' + test_host.encode('ascii') + b'\r\n'
        request += b'Connection: Upgrade\r\n'
        request += b'Upgrade: WebSocket\r\n'
        request += b'Sec-WebSocket-Version: 13\r\n'
        request += b'Sec-WebSocket-Key: ' + nonce + b'\r\n'
        request += b'Sec-WebSocket-Extensions: ' + \
            ext.name.encode('ascii') + b'; ' + \
            offered_params.encode('ascii') + b'\r\n'
        request += b'\r\n'

        ws.receive_bytes(request)
        event = next(ws.events())
        assert isinstance(event, ConnectionRequested)
        ws.accept(event)

        data = ws.bytes_to_send()
        response, headers = data.split(b'\r\n', 1)
        version, code, reason = response.split(b' ')
        headers = parse_headers(headers)

        assert ext.offered == '%s; %s' % (ext.name, offered_params)
        assert headers['sec-websocket-extensions'] == \
            '%s; %s' % (ext.name, ext_params)

    @pytest.mark.parametrize('accept_response', [False, None])
    def test_disinterested_extension_negotiation(self, accept_response):
        test_host = 'frob.nitz'
        test_path = '/fnord'
        ext = FakeExtension(accept_response=accept_response)

        ws = WSConnection(SERVER, extensions=[ext])

        nonce = bytes(random.getrandbits(8) for x in range(0, 16))
        nonce = base64.b64encode(nonce)

        request = b"GET " + test_path.encode('ascii') + b" HTTP/1.1\r\n"
        request += b'Host: ' + test_host.encode('ascii') + b'\r\n'
        request += b'Connection: Upgrade\r\n'
        request += b'Upgrade: WebSocket\r\n'
        request += b'Sec-WebSocket-Version: 13\r\n'
        request += b'Sec-WebSocket-Key: ' + nonce + b'\r\n'
        request += b'Sec-WebSocket-Extensions: ' + \
            ext.name.encode('ascii') + b'\r\n'
        request += b'\r\n'

        ws.receive_bytes(request)
        event = next(ws.events())
        assert isinstance(event, ConnectionRequested)
        ws.accept(event)

        data = ws.bytes_to_send()
        response, headers = data.split(b'\r\n', 1)
        version, code, reason = response.split(b' ')
        headers = parse_headers(headers)

        assert ext.offered == ext.name
        assert 'sec-websocket-extensions' not in headers

    def test_no_params_extension_negotiation(self):
        test_host = 'frob.nitz'
        test_path = '/fnord'
        ext = FakeExtension(accept_response='')

        ws = WSConnection(SERVER, extensions=[ext])

        nonce = bytes(random.getrandbits(8) for x in range(0, 16))
        nonce = base64.b64encode(nonce)

        request = b"GET " + test_path.encode('ascii') + b" HTTP/1.1\r\n"
        request += b'Host: ' + test_host.encode('ascii') + b'\r\n'
        request += b'Connection: Upgrade\r\n'
        request += b'Upgrade: WebSocket\r\n'
        request += b'Sec-WebSocket-Version: 13\r\n'
        request += b'Sec-WebSocket-Key: ' + nonce + b'\r\n'
        request += b'Sec-WebSocket-Extensions: ' + \
            ext.name.encode('ascii') + b'\r\n'
        request += b'\r\n'

        ws.receive_bytes(request)
        event = next(ws.events())
        assert isinstance(event, ConnectionRequested)
        ws.accept(event)

        data = ws.bytes_to_send()
        response, headers = data.split(b'\r\n', 1)
        version, code, reason = response.split(b' ')
        headers = parse_headers(headers)

        assert ext.offered == ext.name
        assert 'sec-websocket-extensions' in headers

    def test_unwanted_extension_negotiation(self):
        test_host = 'frob.nitz'
        test_path = '/fnord'
        ext = FakeExtension(accept_response=False)

        ws = WSConnection(SERVER, extensions=[ext])

        nonce = bytes(random.getrandbits(8) for x in range(0, 16))
        nonce = base64.b64encode(nonce)

        request = b"GET " + test_path.encode('ascii') + b" HTTP/1.1\r\n"
        request += b'Host: ' + test_host.encode('ascii') + b'\r\n'
        request += b'Connection: Upgrade\r\n'
        request += b'Upgrade: WebSocket\r\n'
        request += b'Sec-WebSocket-Version: 13\r\n'
        request += b'Sec-WebSocket-Key: ' + nonce + b'\r\n'
        request += b'Sec-WebSocket-Extensions: pretend\r\n'
        request += b'\r\n'

        ws.receive_bytes(request)
        event = next(ws.events())
        assert isinstance(event, ConnectionRequested)
        ws.accept(event)

        data = ws.bytes_to_send()
        response, headers = data.split(b'\r\n', 1)
        version, code, reason = response.split(b' ')
        headers = parse_headers(headers)

        assert 'sec-websocket-extensions' not in headers

    def test_not_an_http_request_at_all(self):
        ws = WSConnection(SERVER)

        request = b'<xml>Good god, what is this?</xml>\r\n\r\n'

        ws.receive_bytes(request)
        assert isinstance(next(ws.events()), ConnectionFailed)

    def test_h11_somehow_loses_its_mind(self):
        ws = WSConnection(SERVER)
        ws._upgrade_connection.next_event = lambda: object()

        ws.receive_bytes(b'')
        assert isinstance(next(ws.events()), ConnectionFailed)



# -*- coding: utf-8 -*-
"""
wsproto/utilities
~~~~~~~~~~~~~~~~~

Utility functions that do not belong in a separate module.
"""

# Some convenience utilities for working with HTTP headers
def normed_header_dict(h11_headers):
    # This mangles Set-Cookie headers. But it happens that we don't care about
    # any of those, so it's OK. For every other HTTP header, if there are
    # multiple instances then you're allowed to join them together with
    # commas.
    name_to_values = {}
    for name, value in h11_headers:
        name_to_values.setdefault(name, []).append(value)
    name_to_normed_value = {}
    for name, values in name_to_values.items():
        name_to_normed_value[name] = b", ".join(values)
    return name_to_normed_value


# We use this for parsing the proposed protocol list, and for parsing the
# proposed and accepted extension lists. For the proposed protocol list it's
# fine, because the ABNF is just 1#token. But for the extension lists, it's
# wrong, because those can contain quoted strings, which can in turn contain
# commas. XX FIXME
def split_comma_header(value):
    return [piece.decode('ascii').strip() for piece in value.split(b',')]



#!/usr/bin/env python

import os
import re
import sys

from setuptools import setup, find_packages

PROJECT_ROOT = os.path.dirname(__file__)

with open(os.path.join(PROJECT_ROOT, 'README.rst')) as file_:
    long_description = file_.read()

# Get the version
version_regex = r'__version__ = ["\']([^"\']*)["\']'
with open(os.path.join(PROJECT_ROOT, 'wsproto/__init__.py')) as file_:
    text = file_.read()
    match = re.search(version_regex, text)

    if match:
        version = match.group(1)
    else:
        raise RuntimeError("No version number found!")

# Stealing this from Cory Benfield who stole it from Kenneth Reitz
if sys.argv[-1] == 'publish':
    os.system('python setup.py sdist upload')
    sys.exit()

setup(
    name='wsproto',
    version=version,
    description='WebSockets state-machine based protocol implementation',
    long_description=long_description,
    author='Benno Rice',
    author_email='benno@jeamland.net',
    url='https://github.com/python-hyper/wsproto/',
    packages=find_packages(),
    package_data={'': ['LICENSE', 'README.rst']},
    package_dir={'wsproto': 'wsproto'},
    include_package_data=True,
    license='MIT',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: Implementation :: CPython',
        'Programming Language :: Python :: Implementation :: PyPy',
    ],
    install_requires=[
        'h11 ~= 0.8.1',  # means: 0.8.x where x >= 1
    ],
    extras_require={
        ':python_version == "2.7"':
            ['enum34>=1.0.4, <2'],
    }
)



import wsproto.extensions as wpext
import wsproto.frame_protocol as fp


class TestExtension(object):
    def test_enabled(self):
        ext = wpext.Extension()
        assert not ext.enabled()

    def test_offer(self):
        ext = wpext.Extension()
        assert ext.offer(None) is None

    def test_accept(self):
        ext = wpext.Extension()
        assert ext.accept(None, None) is None

    def test_finalize(self):
        ext = wpext.Extension()
        assert ext.finalize(None, None) is None

    def test_frame_inbound_header(self):
        ext = wpext.Extension()
        result = ext.frame_inbound_header(None, None, None, None)
        assert result == fp.RsvBits(False, False, False)

    def test_frame_inbound_payload_data(self):
        ext = wpext.Extension()
        data = object()
        assert ext.frame_inbound_payload_data(None, data) == data

    def test_frame_inbound_complete(self):
        ext = wpext.Extension()
        assert ext.frame_inbound_complete(None, None) is None

    def test_frame_outbound(self):
        ext = wpext.Extension()
        rsv = fp.RsvBits(True, True, True)
        data = object()
        assert ext.frame_outbound(None, None, rsv, data, None) == (rsv, data)



import pytest

from h11 import Request

from wsproto.events import (
    ConnectionClosed,
    ConnectionEstablished,
    ConnectionRequested,
)
from wsproto.frame_protocol import CloseReason


def test_connection_requested_repr_no_subprotocol():
    method = b'GET'
    target = b'/foo'
    headers = {
        b'host': b'localhost',
        b'sec-websocket-version': b'13',
    }
    http_version = b'1.1'

    req = Request(method=method, target=target, headers=list(headers.items()),
                  http_version=http_version)

    event = ConnectionRequested([], req)
    r = repr(event)

    assert 'ConnectionRequested' in r
    assert target.decode('ascii') in r


def test_connection_requested_repr_with_subprotocol():
    method = b'GET'
    target = b'/foo'
    headers = {
        b'host': b'localhost',
        b'sec-websocket-version': b'13',
        b'sec-websocket-protocol': b'fnord',
    }
    http_version = b'1.1'

    req = Request(method=method, target=target, headers=list(headers.items()),
                  http_version=http_version)

    event = ConnectionRequested([], req)
    r = repr(event)

    assert 'ConnectionRequested' in r
    assert target.decode('ascii') in r
    assert headers[b'sec-websocket-protocol'].decode('ascii') in r


@pytest.mark.parametrize('subprotocol,extensions', [
    ('sproto', None),
    (None, ['fake']),
    ('sprout', ['pretend']),
])
def test_connection_established_repr(subprotocol, extensions):
    event = ConnectionEstablished(subprotocol, extensions)
    r = repr(event)

    if subprotocol:
        assert subprotocol in r
    if extensions:
        for extension in extensions:
            assert extension in r


@pytest.mark.parametrize('code,reason', [
    (CloseReason.NORMAL_CLOSURE, None),
    (CloseReason.NORMAL_CLOSURE, 'because i felt like it'),
    (CloseReason.INVALID_FRAME_PAYLOAD_DATA, 'GOOD GOD WHAT DID YOU DO'),
])
def test_connection_closed_repr(code, reason):
    event = ConnectionClosed(code, reason)
    r = repr(event)

    assert repr(code) in r
    if reason:
        assert reason in r
